"""The Driving Continuous Environment."""
import math
from itertools import product
from typing import Any, Dict, List, NamedTuple, Optional, Set, Tuple, Union, cast

import numpy as np
from gymnasium import spaces
from pymunk import Vec2d

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.continuous.core import (
    AGENT_COLORS,
    CollisionType,
    Coord,
    FloatCoord,
    PMBodyState,
    SquareContinuousWorld,
    clip_actions,
    generate_interior_walls,
)
from posggym.utils import seeding


class VehicleState(NamedTuple):
    """The state of a vehicle in the Driving Environment."""

    body: np.ndarray
    dest_coord: np.ndarray
    status: np.ndarray
    min_dest_dist: np.ndarray


DState = Tuple[VehicleState, ...]

# Obs = (sensor obs, dir, vx, vy, dest_coord)
DObs = np.ndarray
# angular velocity, linear velocity
DAction = np.ndarray


class DrivingContinuousEnv(DefaultEnv[DState, DObs, DAction]):
    """The Driving Continuous World Environment.

    A general-sum 2D continuous world problem involving multiple agents. Each agent
    controls a vehicle and is tasked with driving the vehicle from it's start
    location to a destination location while avoiding crashing into other vehicles.
    This requires agents to coordinate to avoid collisions and can be used to explore
    conventions in multi-agent decision-making.

    Possible Agents
    ---------------
    The environment supports two or more agents, depending on the world layout. It is
    possible for some agents to finish the episode before other agents by either
    crashing or reaching their destination, and so not all agents are guaranteed to be
    active at the same time. All agents will be active at the start of the episode.

    State Space
    -----------
    Each state is made up of the state of each vehicle (see `VehicleState` class),
    which in turn is defined by the vehicle's:

    - `(x, y)` coordinates in [0, world_size]
    - direction in [-2π, 2π]
    - x, y velocity both in [-1, 1]
    - the angular velocity of the vehicle in [-2π, 2π]
    - the `(x, y)` coordinate of the vehicles destination
    - whether the vehicle has reached it's destination or not: `1` or `0`
    - whether the vehicle has crashed or not: `1` or `0`
    - the minimum distance to the destination achieved by the vehicle in the current
      episode, if the environment was discrete.

    Action Space
    ------------
    Each agent has 2 actions, which are the angular velocity and linear acceleration.
    Each agent's actions is made up of two parts. The first action component specifies
    the angular velocity in `[-pi/4, pi/4]`, and the second component specifies the
    linear acceleration in `[-0.25, 0.25]`.

    Observation Space
    -----------------
    Each agent observes a local circle around themselves as a vector. This is achieved
    by a series of 'n_sensors' lines starting at the agent which extend for a distance
    of 'obs_dist'. For each line the agent observes the closest entity along the line,
    specifically if there is a wall or another vehicle. Along with the sensor reading
    each agent also observes their vehicles angle, velocity (in x, y), and the distance
    to their destination.

    This table enumerates the observation space:

    | Index: start          | Description                          |  Values   |
    | :-------------------: | :----------------------------------: | :-------: |
    | 0                     | Wall distance                        | [0, d]    |
    | n_sensors             | Other vehicle distance               | [0, d]    |
    | 2 * n_sensors         | Vehicle angle                        | [-2π, 2π] |
    | 2 * n_sensors + 1     | Vehicle x velocity                   | [-1, 1]   |
    | 2 * n_sensors + 2     | Vehicle y velocity                   | [-1, 1]   |
    | 2 * n_sensors + 3     | distance to destination along x axis | [0, s]    |
    | 2 * n_sensors + 4     | distance to destination along y axis | [0, s]    |

    Where `d = obs_dist` and `s = world.size`

    If an entity is not observed by a sensor (i.e. it's not within `obs_dist` or is not
    the closest entity to the observing agent along the line), The distance reading will
    be `obs_dist`.

    The sensor reading ordering is relative to the agent's direction. I.e. the values
    for the first sensor at indices `0`, `n_sensors`, correspond to the distance
    reading to a wall, and other vehicle, respectively, in the direction the agent is
    facing.

    Rewards
    -------
    All agents receive a penalty of `0.0` for each step. They receive a penalty of
    `-1.0` for crashing (i.e. hitting another vehicle). A reward of `1.0` is given if
    the agent reaches it's destination and a reward of `0.05` is given to the agent at
    certain points as it makes progress towards it's destination (i.e. as it reduces
    it's minimum distance achieved along the shortest path to the destination for the
    episode).

    Dynamics
    --------
    Actions are deterministic and movement is determined by direction the vehicle is
    facing and it's speed. Vehicles are able to reverse, but cannot change direction
    while reversing.

    Max and min velocity are `1.0` and `-1.0`, and max linear acceleration is `0.25`,
    while max angular velocity is `π / 4`.

    Starting State
    --------------
    Each agent is randomly assigned to one of the possible starting locations in the
    world and one of the possible destination locations, with no two agents starting in
    the same location or having the same destination location. The possible start and
    destination locations are determined by the world layout being used.

    Episodes End
    ------------
    Episodes end when all agents have either reached their destination or crashed. By
    default a `max_episode_steps` is also set for each DrivingContinuous environment.
    The default value is `200` steps, but this may need to be adjusted when using
    larger worlds (this can be done by manually specifying a value for
    `max_episode_steps` when creating the environment with `posggym.make`).

    Arguments
    ---------

    - `world` - the world layout to use. This can either be a string specifying one of
         the supported worlds, or a custom :class:`DrivingWorld` object
         (default = `"14x14RoundAbout"`).
    - `num_agents` - the number of agents in the environment (default = `2`).
    - `obs_dist` - the sensor observation distance, specifying the distance away from
         itself which an agent can observe along each sensor (default = `5.0`).
    - `n_sensors` - the number of sensor lines eminating from the agent. The agent will
         observe at `n_sensors` equidistance intervals over `[0, 2*pi]`
         (default = `16`).

    Available variants
    ------------------
    The DrivingContinuous environment comes with a number of pre-built world layouts
    which can be passed as an argument to `posggym.make`, to create different worlds:

    | World name        | Max number of agents | World size |
    |-------------------|----------------------|----------- |
    | `6x6`             | 6                    | 6x6        |
    | `7x7Blocks`       | 4                    | 7x7        |
    | `7x7CrissCross`   | 6                    | 7x7        |
    | `7x7RoundAbout`   | 4                    | 7x7        |
    | `14x14Blocks`     | 4                    | 14x14      |
    | `14x14CrissCross` | 8                    | 14x14      |
    | `14x14RoundAbout` | 4                    | 14x14      |


    For example to use the DrivingContinuous environment with the `7x7RoundAbout`
    layout and 2 agents, you would use:

    ```python
    import posggym
    env = posggym.make('DrivingContinuous-v0', world="7x7RoundAbout", num_agents=2)
    ```

    Version History
    ---------------
    - `v0`: Initial version

    References
    ----------
    - Adam Lerer and Alexander Peysakhovich. 2019. Learning Existing Social Conventions
    via Observationally Augmented Self-Play. In Proceedings of the 2019 AAAI/ACM
    Conference on AI, Ethics, and Society. 107–114.
    - Kevin R. McKee, Joel Z. Leibo, Charlie Beattie, and Richard Everett. 2022.
    Quantifying the Effects of Environment and Population Diversity in Multi-Agent
    Reinforcement Learning. Autonomous Agents and Multi-Agent Systems 36, 1 (2022), 1–16

    """

    metadata = {
        "render_modes": ["human", "rgb_array"],
        "render_fps": 15,
    }

    def __init__(
        self,
        world: Union[str, "DrivingWorld"] = "14x14RoundAbout",
        num_agents: int = 2,
        obs_dist: float = 5.0,
        n_sensors: int = 16,
        render_mode: Optional[str] = None,
    ):
        super().__init__(
            DrivingContinuousModel(world, num_agents, obs_dist, n_sensors),
            render_mode=render_mode,
        )
        self.window_surface = None
        self.clock = None
        self.window_size = 600
        self.draw_options = None
        self.world = None
        self.blocked_surface = None

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        return self._render_img()

    def _render_img(self):
        # import posggym.envs.continuous.render as render_lib
        import pygame
        from pymunk import Transform, pygame_util

        model = cast(DrivingContinuousModel, self.model)
        state = cast(DState, self.state)
        scale_factor = self.window_size / model.world.size

        if self.window_surface is None:
            pygame.init()
            if self.render_mode == "human":
                pygame.display.init()
                pygame.display.set_caption(self.__class__.__name__)
                self.window_surface = pygame.display.set_mode(
                    (self.window_size, self.window_size)
                )
            else:
                self.window_surface = pygame.Surface(
                    (self.window_size, self.window_size)
                )
            # Turn off alpha since we don't use it.
            self.window_surface.set_alpha(None)

        if self.clock is None:
            self.clock = pygame.time.Clock()

        if self.draw_options is None:
            pygame_util.positive_y_is_up = False
            self.draw_options = pygame_util.DrawOptions(self.window_surface)
            self.draw_options.transform = Transform.scaling(scale_factor)
            # don't render collision lines
            self.draw_options.flags = (
                pygame_util.DrawOptions.DRAW_SHAPES
                | pygame_util.DrawOptions.DRAW_CONSTRAINTS
            )

        if self.world is None:
            # get copy of model world, so we can use it for rendering without
            # affecting the original
            self.world = model.world.copy()

        if self.blocked_surface is None:
            self.blocked_surface = pygame.Surface((self.window_size, self.window_size))
            self.blocked_surface.fill(pygame.Color("white"))
            cell_size = scale_factor
            for coord in self.world.blocked_coords:
                rect = (
                    round(coord[0] * cell_size),
                    round(coord[1] * cell_size),
                    round(cell_size),
                    round(cell_size),
                )
                self.blocked_surface.fill(self.world.BLOCK_COLOR, rect=rect)

        for i, v_state in enumerate(state):
            self.world.set_entity_state(f"vehicle_{i}", v_state.body)

        # Need to do this for space to update with changes
        self.world.space.step(0.0001)

        # reset screen
        self.window_surface.fill(pygame.Color("white"))

        # add blocks
        self.window_surface.blit(self.blocked_surface, (0, 0))

        lines_colors = ["red", "green", "black"]
        # draw sensor lines
        n_sensors = model.n_sensors
        for i, obs_i in self._last_obs.items():
            line_obs = obs_i[: model.sensor_obs_dim]
            x, y, agent_angle = state[int(i)].body[:3]
            angle_inc = 2 * math.pi / n_sensors
            for k in range(n_sensors):
                values = [
                    line_obs[k],
                    line_obs[n_sensors + k],
                ]
                dist_idx = min(range(len(values)), key=values.__getitem__)
                dist = values[dist_idx]
                dist_idx = len(values) if dist == model.obs_dist else dist_idx

                angle = angle_inc * k + agent_angle
                end_x = x + dist * math.cos(angle)
                end_y = y + dist * math.sin(angle)
                scaled_start = (int(x * scale_factor), int(y * scale_factor))
                scaled_end = int(end_x * scale_factor), (end_y * scale_factor)

                pygame.draw.line(
                    self.window_surface,
                    pygame.Color(lines_colors[dist_idx]),
                    scaled_start,
                    scaled_end,
                )

        for i, v in enumerate(state):
            x, y = v.dest_coord[:2]
            scaled_center = (int(x * scale_factor), int(y * scale_factor))
            scaled_r = int(model.world.agent_radius * scale_factor)
            pygame.draw.circle(
                self.window_surface,
                AGENT_COLORS[i][1],
                scaled_center,
                scaled_r,
            )

        self.world.space.debug_draw(self.draw_options)

        if self.render_mode == "human":
            pygame.event.pump()
            pygame.display.update()
            self.clock.tick(self.metadata["render_fps"])
            return None

        return np.transpose(
            np.array(pygame.surfarray.pixels3d(self.window_surface)), axes=(1, 0, 2)
        )

    def close(self) -> None:
        if self.window_surface is not None:
            import pygame

            pygame.display.quit()
            pygame.quit()


class DrivingContinuousModel(M.POSGModel[DState, DObs, DAction]):
    """Driving Problem Model.

    Parameters
    ----------
    world : str, DrivingWorld
        the world environment for the model scenario
    num_agents : int
        the number of agents in the model scenario
    obs_dists : float
        number of cells in front, behind, and to the side that each agent
        can observe
    n_sensors : the number of sensor lines eminating from the agent. The agent will
        observe at `n_sensors` equidistance intervals over `[0, 2*pi]`

    """

    R_STEP_COST = 0.00
    R_CRASH_VEHICLE = -1.0
    R_DESTINATION_REACHED = 1.0
    R_PROGRESS = 0.05

    def __init__(
        self,
        world: Union[str, "DrivingWorld"],
        num_agents: int,
        obs_dist: float,
        n_sensors: int,
    ):
        if isinstance(world, str):
            assert world in SUPPORTED_WORLDS, (
                f"Unsupported world '{world}'. If world argument is a string it must "
                f"be one of: {SUPPORTED_WORLDS.keys()}."
            )
            world_info = SUPPORTED_WORLDS[world]
            world = parseworld_str(
                world_info["world_str"], world_info["supported_num_agents"]
            )
        assert 0 < num_agents <= world.supported_num_agents, (
            f"Supplied DrivingWorld `{world}` does not support {num_agents} "
            "agents. The supported number of agents is from 1 up to "
            f"{world.supported_num_agents}."
        )

        self.world = world
        self.n_sensors = n_sensors
        self.obs_dist = obs_dist
        self.vehicle_collision_dist = 2.1 * self.world.agent_radius

        self.possible_agents = tuple(str(i) for i in range(num_agents))
        self.state_space = spaces.Tuple(
            tuple(
                spaces.Tuple(
                    (
                        # State of vehicle
                        PMBodyState.get_space(self.world.size),
                        # vehicle destination coord
                        spaces.Box(low=0.0, high=self.world.size, shape=(2,)),
                        # dest reached, crashed
                        spaces.MultiBinary(2),
                        # min distance to destination
                        # set this to upper bound of min shortest path distance, so
                        # state space works for generated worlds as well
                        spaces.Box(
                            low=np.array([0], dtype=np.float32),
                            high=np.array([self.world.size**2], dtype=np.float32),
                        ),
                    )
                )
                for _ in range(len(self.possible_agents))
            )
        )

        self.dyaw_limit = math.pi / 4
        self.dvel_limit = 0.25
        self.vel_limit_norm = 1.0
        # dyaw, dvel
        self.action_spaces = {
            i: spaces.Box(
                low=np.array([-self.dyaw_limit, -self.dvel_limit], dtype=np.float32),
                high=np.array([self.dyaw_limit, self.dvel_limit], dtype=np.float32),
            )
            for i in self.possible_agents
        }

        # Observes entity and distance to entity along a n_sensors rays from the agent
        # 0 to n_sensors = wall distance obs
        # n_sensors to (2 * n_sensors) = other vehicle dist
        # Also observs angle, vx, vy, dest dx, desy dy
        self.sensor_obs_dim = self.n_sensors * 2
        self.obs_dim = self.sensor_obs_dim + 5
        sensor_low = [0.0] * self.sensor_obs_dim
        sensor_high = [self.obs_dist] * self.sensor_obs_dim
        self.observation_spaces = {
            i: spaces.Box(
                low=np.array(
                    [*sensor_low, -2 * math.pi, -1, -1, 0, 0], dtype=np.float32
                ),
                high=np.array(
                    [*sensor_high, 2 * math.pi, 1, 1, self.world.size, self.world.size],
                    dtype=np.float32,
                ),
                dtype=np.float32,
            )
            for i in self.possible_agents
        }

        # Add physical entities to the world
        for i in range(len(self.possible_agents)):
            color_i, _ = AGENT_COLORS[i]
            self.world.add_entity(f"vehicle_{i}", None, color=color_i)

        self.is_symmetric = True

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        return {
            i: (self.R_CRASH_VEHICLE, self.R_DESTINATION_REACHED)
            for i in self.possible_agents
        }

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: DState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> DState:
        state = []
        chosen_start_coords: Set[FloatCoord] = set()
        chosen_dest_coords: Set[FloatCoord] = set()
        for i in range(len(self.possible_agents)):
            start_coords_i = self.world.start_coords[i]
            avail_start_coords = start_coords_i.difference(chosen_start_coords)
            start_coord = self.rng.choice(list(avail_start_coords))
            chosen_start_coords.add(start_coord)

            dest_coords_i = self.world.dest_coords[i]
            avail_dest_coords = dest_coords_i.difference(chosen_dest_coords)
            if start_coord in avail_dest_coords:
                avail_dest_coords.remove(start_coord)

            body_state = np.zeros((PMBodyState.num_features()), dtype=np.float32)
            body_state[:2] = start_coord

            _dest_coord = self.rng.choice(list(avail_dest_coords))
            chosen_dest_coords.add(_dest_coord)
            dest_coord = np.array(_dest_coord, dtype=np.float32)

            dest_dist = self.world.get_shortest_path_distance(start_coord, _dest_coord)

            state_i = VehicleState(
                body=body_state,
                dest_coord=dest_coord,
                status=np.array([int(False), int(False)], dtype=np.int8),
                min_dest_dist=np.array([dest_dist], dtype=np.float32),
            )
            state.append(state_i)

        return tuple(state)

    def sample_initial_obs(self, state: DState) -> Dict[str, DObs]:
        return self._get_obs(state)

    def step(
        self, state: DState, actions: Dict[str, DAction]
    ) -> M.JointTimestep[DState, DObs]:
        clipped_actions = clip_actions(actions, self.action_spaces)

        next_state, collision_types = self._get_next_state(state, clipped_actions)
        obs = self._get_obs(next_state)
        rewards = self._get_rewards(state, next_state, collision_types)
        terminated = {i: any(next_state[int(i)].status) for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        all_done = all(terminated.values())

        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        for idx in range(len(self.possible_agents)):
            if next_state[idx].status[0]:
                outcome_i = M.Outcome.WIN
            elif next_state[idx].status[1]:
                outcome_i = M.Outcome.LOSS
            else:
                outcome_i = M.Outcome.NA
            info[str(idx)]["outcome"] = outcome_i

        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(
        self, state: DState, actions: Dict[str, DAction]
    ) -> Tuple[DState, List[CollisionType]]:
        for i in range(len(self.possible_agents)):
            state_i = state[i]
            self.world.set_entity_state(f"vehicle_{i}", state_i.body)

            if state[i].status[0] or state[i].status[1]:
                self.world.update_entity_state(f"vehicle_{i}", vel=(0.0, 0.0))
                continue

            action_i = actions[str(i)]
            v_angle = state_i.body[2] + action_i[0]
            v_vel = Vec2d(*state_i.body[3:5]).rotated(action_i[0]) + (
                action_i[1] * Vec2d(1, 0).rotated(v_angle)
            )
            self.world.update_entity_state(
                f"vehicle_{i}",
                angle=v_angle,
                vel=self.world.clamp_norm(v_vel[0], v_vel[1], self.vel_limit_norm),
            )

        self.world.simulate(1.0 / 10, 10)

        collision_types = [CollisionType.NONE] * len(self.possible_agents)
        new_state: List[Optional[VehicleState]] = [None] * len(self.possible_agents)
        for idx in range(len(self.possible_agents)):
            next_v_body_state = np.array(
                self.world.get_entity_state(f"vehicle_{idx}"), dtype=np.float32
            )

            next_v_body_state[2] = self.world.convert_angle_to_0_2pi_interval(
                next_v_body_state[2]
            )

            # ensure vx, vy is in [-1, 1]
            # with collisions, etc pymunk can sometime push it over this limit
            next_v_body_state[3] = max(-1.0, min(1.0, next_v_body_state[3]))
            next_v_body_state[4] = max(-1.0, min(1.0, next_v_body_state[4]))

            state_i = state[idx]
            next_v_coords = next_v_body_state[:2]
            dest_distance = np.linalg.norm(state_i.dest_coord - next_v_coords)
            crashed = False

            for other_idx, other_v_state in enumerate(new_state):
                if other_v_state is None:
                    continue
                dist = np.linalg.norm(other_v_state.body[:2] - next_v_coords)
                if dist <= self.vehicle_collision_dist:
                    crashed = True
                    collision_types[idx] = CollisionType.AGENT
                    if not other_v_state.status[0]:
                        # must also update other vehicle but only update other vehicle
                        # if it has not already reached destination
                        other_v_state.status[1] = int(True)
                        collision_types[other_idx] = CollisionType.AGENT

            crashed = crashed or bool(state_i.status[1])

            min_dest_dist = min(
                state_i.min_dest_dist[0],
                self.world.get_shortest_path_distance(
                    (next_v_body_state[0], next_v_body_state[1]),
                    (state_i.dest_coord[0], state_i.dest_coord[1]),
                ),
            )

            new_state[idx] = VehicleState(
                body=next_v_body_state,
                dest_coord=state_i.dest_coord,
                status=np.array(
                    [int(dest_distance <= self.world.agent_radius), int(crashed)],
                    dtype=np.int8,
                ),
                min_dest_dist=np.array([min_dest_dist], dtype=np.float32),
            )

        final_state = [ns for ns in new_state if ns is not None]
        assert len(final_state) == len(new_state)

        return tuple(final_state), collision_types

    def _get_obs(self, state: DState) -> Dict[str, DObs]:
        return {i: self._get_agent_obs(i, state) for i in self.possible_agents}

    def _get_agent_obs(self, agent_id: str, state: DState) -> np.ndarray:
        state_i = state[int(agent_id)]
        if state_i.status[0] or state_i.status[1]:
            return np.zeros((self.obs_dim,), dtype=np.float32)

        pos_i = (state_i.body[0], state_i.body[1], state_i.body[2])
        vehicle_coords = np.array(
            [[s.body[0], s.body[1]] for i, s in enumerate(state) if i != int(agent_id)]
        )

        ray_dists, ray_col_type = self.world.check_collision_circular_rays(
            pos_i,
            ray_distance=self.obs_dist,
            n_rays=self.n_sensors,
            other_agents=vehicle_coords,
            include_blocks=True,
            check_walls=True,
            use_relative_angle=True,
        )

        obs = np.full((self.obs_dim,), self.obs_dist, dtype=np.float32)
        # Can bucket NONE type collisions with block/border/wall collisions, since these
        # will have distance of obs_dist and hence not affect final obs
        obs_type_idx = np.where(ray_col_type == CollisionType.AGENT.value, 1, 0)
        flat_obs_idx = np.ravel_multi_index(
            (obs_type_idx, np.arange(self.n_sensors)), dims=(2, self.n_sensors)
        )
        obs[flat_obs_idx] = ray_dists

        d = self.sensor_obs_dim
        obs[d] = self.world.convert_angle_to_0_2pi_interval(state_i.body[2])
        obs[d + 1] = max(-1.0, min(1.0, state_i.body[3]))
        obs[d + 2] = max(-1.0, min(1.0, state_i.body[4]))
        obs[d + 3] = abs(state_i.dest_coord[0] - pos_i[0])
        obs[d + 4] = abs(state_i.dest_coord[1] - pos_i[1])

        return obs

    def _get_rewards(
        self, state: DState, next_state: DState, collision_types: List[CollisionType]
    ) -> Dict[str, float]:
        rewards: Dict[str, float] = {}
        for i in self.possible_agents:
            idx = int(i)
            if any(state[idx].status):
                # already in terminal/rewarded state
                r_i = 0.0
            elif collision_types[idx] == CollisionType.AGENT:
                # Treat as if crashed into a vehicle
                r_i = self.R_CRASH_VEHICLE
            elif next_state[idx].status[0]:
                r_i = self.R_DESTINATION_REACHED
            else:
                r_i = self.R_STEP_COST

            progress = (state[idx].min_dest_dist - next_state[idx].min_dest_dist)[0]
            r_i += max(0, progress) * self.R_PROGRESS
            rewards[i] = r_i
        return rewards


class DrivingWorld(SquareContinuousWorld):
    """A world for the Driving Problem."""

    def __init__(
        self,
        size: int,
        blocked_coords: Set[Coord],
        start_coords: List[Set[FloatCoord]],
        dest_coords: List[Set[FloatCoord]],
    ):
        interior_walls = generate_interior_walls(size, size, blocked_coords)
        super().__init__(
            size=size,
            blocks=None,
            interior_walls=interior_walls,
            agent_radius=0.4,
            border_thickness=0.01,
            enable_agent_collisions=True,
        )
        assert len(start_coords) == len(dest_coords)
        self._blocked_coords = blocked_coords
        self.start_coords = start_coords
        self.dest_coords = dest_coords
        self.shortest_paths = self.get_all_shortest_paths(
            set.union(*dest_coords)  # type: ignore
        )

    def copy(self) -> "DrivingWorld":
        assert self._blocked_coords is not None
        world = DrivingWorld(
            size=int(self.size),
            blocked_coords=self._blocked_coords,
            start_coords=self.start_coords,
            dest_coords=self.dest_coords,
        )
        for id, (body, shape) in self.entities.items():
            # make copies of each entity, and ensure the copies are linked correctly
            # and added to the new world and world space
            body = body.copy()
            shape = shape.copy()
            shape.body = body
            world.space.add(body, shape)
            world.entities[id] = (body, shape)
        return world

    @property
    def supported_num_agents(self) -> int:
        """Get the number of agents supported by this world."""
        return len(self.start_coords)

    def get_shortest_path_distance(self, coord: FloatCoord, dest: FloatCoord) -> int:
        """Get the shortest path distance from coord to destination."""
        coord_c = self.convert_to_coord(coord)
        dest_c = self.convert_to_coord(dest)
        return int(self.shortest_paths[dest_c][coord_c])

    def get_max_shortest_path_distance(self) -> int:
        """Get the longest shortest path distance to any destination."""
        return int(max([max(d.values()) for d in self.shortest_paths.values()]))


def parseworld_str(world_str: str, supported_num_agents: int) -> DrivingWorld:
    """Parse a str representation of a world.

    Notes on world str representation:

    . = empty/unallocated cell
    # = a block
    0, 1, ..., 9 = starting location for agent with given index
    + = starting point for any agent
    a, b, ..., j = destination location for agent with given index
                   (a=0, b=1, ..., j=9)
    - = destination location for any agent

    Examples (" " quotes and newline chars omitted):

    1. A 3x3 world with two agents, one block, and where each agent has a single
    starting location and a single destination location.

    a1.
    .#.
    .0.

    2. A 6x6 world with 4 common start and destination locations and many
       blocks. This world can support up to 4 agents.

    +.##+#
    ..+#.+
    #.###.
    #.....
    -..##.
    #-.-.-

    """
    row_strs = world_str.splitlines()
    assert len(row_strs) > 1
    assert all(len(row) == len(row_strs[0]) for row in row_strs)
    assert len(row_strs[0]) > 1

    height = len(row_strs)
    width = len(row_strs[0])

    agent_start_chars = set(["+"] + [str(i) for i in range(10)])
    agent_dest_chars = set(["-"] + list("abcdefghij"))

    blocked_coords: Set[Coord] = set()
    shared_start_coords: Set[FloatCoord] = set()
    agent_start_coords_map: Dict[int, Set[FloatCoord]] = {}
    shared_dest_coords: Set[FloatCoord] = set()
    agent_dest_coords_map: Dict[int, Set[FloatCoord]] = {}
    for r, c in product(range(height), range(width)):
        coord = (c + 0.5, r + 0.5)
        char = row_strs[r][c]

        if char == "#":
            blocked_coords.add((c, r))
        elif char in agent_start_chars:
            if char != "+":
                agent_id = int(char)
                if agent_id not in agent_start_coords_map:
                    agent_start_coords_map[agent_id] = set()
                agent_start_coords_map[agent_id].add(coord)
            else:
                shared_start_coords.add(coord)
        elif char in agent_dest_chars:
            if char != "-":
                agent_id = ord(char) - ord("a")
                if agent_id not in agent_dest_coords_map:
                    agent_dest_coords_map[agent_id] = set()
                agent_dest_coords_map[agent_id].add(coord)
            else:
                shared_dest_coords.add(coord)

    assert (
        len(shared_start_coords) + len(agent_start_coords_map) >= supported_num_agents
    )
    assert len(shared_dest_coords) + len(agent_dest_coords_map) >= supported_num_agents

    included_agent_ids = list({*agent_start_coords_map, *agent_dest_coords_map})
    if len(included_agent_ids) > 0:
        assert max(included_agent_ids) < supported_num_agents

    start_coords: List[Set[FloatCoord]] = []
    dest_coords: List[Set[FloatCoord]] = []
    for i in range(supported_num_agents):
        agent_start_coords = set(shared_start_coords)
        agent_start_coords.update(agent_start_coords_map.get(i, {}))
        start_coords.append(agent_start_coords)

        agent_dest_coords = set(shared_dest_coords)
        agent_dest_coords.update(agent_dest_coords_map.get(i, {}))
        dest_coords.append(agent_dest_coords)

    return DrivingWorld(
        size=width,
        blocked_coords=blocked_coords,
        start_coords=start_coords,
        dest_coords=dest_coords,
    )


SUPPORTED_WORLDS: Dict[str, Dict[str, Any]] = {
    "6x6Intersection": {
        "world_str": (
            "##0b##\n" "##..##\n" "d....3\n" "2....c\n" "##..##\n" "##a1##\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 20,
    },
    "7x7Blocks": {
        "world_str": (
            "#-...-#\n"
            "-##.##+\n"
            ".##.##.\n"
            ".......\n"
            ".##.##.\n"
            "-##.##+\n"
            "#+...+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "7x7CrissCross": {
        "world_str": (
            "#-#-#-#\n"
            "-.....+\n"
            "#.#.#.#\n"
            "-.....+\n"
            "#.#.#.#\n"
            "-.....+\n"
            "#+#+#+#\n"
        ),
        "supported_num_agents": 6,
        "max_episode_steps": 50,
    },
    "7x7RoundAbout": {
        "world_str": (
            "#-...-#\n"
            "-##.##+\n"
            ".#...#.\n"
            "...#...\n"
            ".#...#.\n"
            "-##.##+\n"
            "#+...+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "14x14Blocks": {
        "world_str": (
            "#-..........-#\n"
            "-###.####.###+\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "..............\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "..............\n"
            ".###.####.###.\n"
            ".###.####.###.\n"
            "-###.####.###+\n"
            "#+..........+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
    "14x14CrissCross": {
        "world_str": (
            "##-##-##-##-##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##.##.##.##.##\n"
            "-............+\n"
            "##.##.##.##.##\n"
            "##+##+##+##+##\n"
        ),
        "supported_num_agents": 8,
        "max_episode_steps": 50,
    },
    "14x14RoundAbout": {
        "world_str": (
            "#-..........-#\n"
            "-#####..#####+\n"
            ".#####..#####.\n"
            ".#####..#####.\n"
            ".###......###.\n"
            ".###......###.\n"
            "......##......\n"
            "......##......\n"
            ".###......###.\n"
            ".###......###.\n"
            ".#####..#####.\n"
            ".#####..#####.\n"
            "-#####..#####+\n"
            "#+..........+#\n"
        ),
        "supported_num_agents": 4,
        "max_episode_steps": 50,
    },
}



"""Wrapper to enforce the proper ordering of environment operations."""
import posggym
from posggym.error import ResetNeeded


class OrderEnforcing(posggym.Wrapper):
    """Wraps environment to enforce environment is reset before stepped.

    Will produce an error if :meth:`step` is called before an initial :meth:`reset`.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper
    disable_render_order_enforcing : bool
        Whether to disable enforcing of reset before render is called or not.

    Note
    ----
    This implementation is based on the similar Gymnasium wrapper:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/order_enforcing.py

    """

    def __init__(self, env: posggym.Env, disable_render_order_enforcing: bool = False):
        super().__init__(env)
        self._has_reset = False
        self._disable_render_order_enforcing = disable_render_order_enforcing

    def step(self, actions):
        if not self._has_reset:
            raise ResetNeeded("Cannot call env.step() before calling env.reset()")
        return self.env.step(actions)

    def reset(self, **kwargs):
        self._has_reset = True
        return self.env.reset(**kwargs)

    def render(self):
        if not self._disable_render_order_enforcing and not self._has_reset:
            raise ResetNeeded(
                "Cannot call `env.render()` before calling `env.reset()`, if this is an"
                " intended action, set `disable_render_order_enforcing=True` on the "
                "OrderEnforcer wrapper."
            )
        return self.env.render()

    @property
    def has_reset(self) -> bool:
        """Returns if the environment has been reset before."""
        return self._has_reset



"""posggym specific logging.

Adapted from Farama Foundation Gymnasium:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/logger.py

"""
import sys
import warnings
from typing import Optional, Type

from gymnasium.utils.colorize import colorize


DEBUG = 10
INFO = 20
WARN = 30
ERROR = 40
DISABLED = 50

min_level = 30


def set_level(level: int):
    """Set logging threshold on current logger."""
    global min_level
    min_level = level


def debug(msg: str, *args):
    """Log debug message."""
    if min_level <= DEBUG:
        print(f"DEBUG: {msg % args}", file=sys.stderr)


def info(msg: str, *args):
    """Log info message."""
    if min_level <= INFO:
        print(f"INFO: {msg % args}", file=sys.stderr)


def warn(
    msg: str,
    *args: object,
    category: Optional[Type[Warning]] = None,
    stacklevel: int = 1,
):
    """Raises a warning to the user if the min_level <= WARN.

    Arguments
    ---------
    msg: str
        The message to warn the user
    *args: object
        Additional information to warn the user
    category: Type[Warning], Optional
        The category of warning (default=None)
    stacklevel: int, Optional
        The stack level to raise to (default=1)

    """
    if min_level <= WARN:
        warnings.warn(
            colorize(f"WARN: {msg % args}", "yellow"),
            category=category,
            stacklevel=stacklevel + 1,
        )


def deprecation(msg: str, *args: object):
    """Logs a deprecation warning to users."""
    warn(msg, *args, category=DeprecationWarning, stacklevel=2)


def error(msg: str, *args):
    """Log error message."""
    if min_level <= ERROR:
        print(colorize(f"ERROR: {msg % args}", "red"), file=sys.stderr)



"""Registers the internal POSGGym environments.

Based on the Farama Foundation Gymnasium API:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/envs/__init__.py

"""
import math

from posggym.envs.registration import make, pprint_registry, register, registry, spec


# Classic
# -------------------------------------------

register(
    id="MultiAccessBroadcastChannel-v0", entry_point="posggym.envs.classic.mabc:MABCEnv"
)

register(
    id="MultiAgentTiger-v0", entry_point="posggym.envs.classic.tiger:MultiAgentTigerEnv"
)

register(
    id="RockPaperScissors-v0",
    entry_point="posggym.envs.classic.rock_paper_scissors:RockPaperScissorsEnv",
)

# Continuous
# -------------------------------------------
register(
    id="DrivingContinuous-v0",
    entry_point="posggym.envs.continuous.driving_continuous:DrivingContinuousEnv",
    max_episode_steps=100,
    kwargs={
        "world": "14x14RoundAbout",
        "num_agents": 2,
        "obs_dist": 5.0,
        "n_sensors": 16,
    },
)

register(
    id="DroneTeamCapture-v0",
    entry_point="posggym.envs.continuous.drone_team_capture:DroneTeamCaptureEnv",
    max_episode_steps=500,
    kwargs={
        "num_agents": 3,
        "n_communicating_pursuers": None,
        "arena_radius": 430,
        "observation_limit": None,
        "velocity_control": False,
        "capture_radius": 30,
        "use_q_reward": False,
    },
)

register(
    id="PredatorPreyContinuous-v0",
    entry_point="posggym.envs.continuous.predator_prey_continuous:PredatorPreyContinuousEnv",
    max_episode_steps=100,
    kwargs={
        "world": "10x10",
        "num_predators": 2,
        "num_prey": 3,
        "cooperative": True,
        "prey_strength": None,
        "obs_dist": 4,
        "n_sensors": 16,
    },
)

register(
    id="PursuitEvasionContinuous-v0",
    entry_point="posggym.envs.continuous.pursuit_evasion_continuous:PursuitEvasionContinuousEnv",
    max_episode_steps=100,
    kwargs={
        "world": "16x16",
        "max_obs_distance": None,
        "fov": math.pi / 3,
        "n_sensors": 16,
        "normalize_reward": True,
        "use_progress_reward": True,
    },
)

# Grid World
# -------------------------------------------

# Driving
register(
    id="Driving-v0",
    entry_point="posggym.envs.grid_world.driving:DrivingEnv",
    max_episode_steps=50,
    kwargs={
        "grid": "14x14RoundAbout",
        "num_agents": 2,
        "obs_dim": (3, 1, 1),
        "obstacle_collisions": False,
    },
)


# Driving Gen
# Default grid generator params is "14x14"
register(
    id="DrivingGen-v0",
    entry_point="posggym.envs.grid_world.driving_gen:DrivingGenEnv",
    max_episode_steps=50,
    kwargs={
        "num_agents": 2,
        "obs_dim": (3, 1, 1),
        "obstacle_collisions": False,
        "generator_params": "14x14",
        "n_grids": None,
        "shuffle_grid_order": True,
    },
)


# Level-Based Foraging
# Based on github.com/semitable/lb-foraging/
register(
    id="LevelBasedForaging-v2",
    entry_point="posggym.envs.grid_world.level_based_foraging:LevelBasedForagingEnv",
    max_episode_steps=50,
    kwargs={
        "num_agents": 2,
        "max_agent_level": 3,
        "field_size": (10, 10),
        "max_food": 8,
        "sight": 2,
        "force_coop": False,
        "static_layout": False,
        "normalize_reward": True,
        "observation_mode": "tuple",
        "penalty": 0.0,
    },
)


# Predator-Prey
register(
    id="PredatorPrey-v0",
    entry_point="posggym.envs.grid_world.predator_prey:PredatorPreyEnv",
    max_episode_steps=50,
    kwargs={
        "grid": "10x10",
        "num_predators": 2,
        "num_prey": 3,
        "cooperative": True,
        "prey_strength": None,
        "obs_dim": 2,
    },
)


# Pursuit-Evasion
register(
    id="PursuitEvasion-v0",
    entry_point="posggym.envs.grid_world.pursuit_evasion:PursuitEvasionEnv",
    max_episode_steps=100,
    kwargs={
        "grid": "16x16",
        "action_probs": 1.0,
        "max_obs_distance": 12,
        "normalize_reward": True,
        "use_progress_reward": True,
    },
)


# Two Paths
register(
    id="TwoPaths-v0",
    entry_point="posggym.envs.grid_world.two_paths:TwoPathsEnv",
    max_episode_steps=20,
    kwargs={"grid_size": 7, "action_probs": 1.0},
)


# Unmanned Aerial Vehicle (UAV)
register(
    id="UAV-v0",
    entry_point="posggym.envs.grid_world.uav:UAVEnv",
    max_episode_steps=50,
    kwargs={"grid": 5},
)



"""The Pursuit-Evasion Grid World Environment."""
from collections import deque
from typing import Any, Deque, Dict, List, NamedTuple, Optional, Set, Tuple, Union

from gymnasium import spaces

import posggym.model as M
from posggym import logger
from posggym.core import DefaultEnv
from posggym.envs.grid_world.core import Coord, Direction, Grid
from posggym.utils import seeding


# State = (e_coord, e_dir, p_coord, p_dir, e_0_coord, p_0_coord, e_goal_coord)
INITIAL_DIR = Direction.NORTH


class PEState(NamedTuple):
    """Environment state in Pursuit Evastion problem."""

    evader_coord: Coord
    evader_dir: Direction
    pursuer_coord: Coord
    pursuer_dir: Direction
    evader_start_coord: Coord
    pursuer_start_coord: Coord
    evader_goal_coord: Coord
    min_goal_dist: int  # for evader


# Action = Direction of movement
# 0 = Forward, 1 = Backward, 2 = Left, 3 = Right
PEAction = int
ACTION_TO_DIR = [
    # Forward
    [Direction.NORTH, Direction.EAST, Direction.SOUTH, Direction.WEST],
    # Backward
    [Direction.SOUTH, Direction.WEST, Direction.NORTH, Direction.EAST],
    # Left
    [Direction.WEST, Direction.NORTH, Direction.EAST, Direction.SOUTH],
    # Right
    [Direction.EAST, Direction.SOUTH, Direction.WEST, Direction.NORTH],
]

# E Obs = Tuple[WallObs, seen, heard, e_0_coord, p_0_coord, goal_coord]
#       = Tuple[Tuple[int, int, int, int, int, int], Coord, Coord, Coord]
# P Obs = Tuple[WallObs, seen, heard, e_0_coord, p_0_coord, blank_coord]
#       = Tuple[Tuple[int, int, int, int, int, int], Coord, Coord, Coord]
# Note, we use blank_coord for P Obs so Obs spaces are identical between the
# two agents. The blank_coord is always (0, 0).
PEEvaderObs = Tuple[Tuple[int, ...], Coord, Coord, Coord]
PEPursuerObs = Tuple[Tuple[int, ...], Coord, Coord, Coord]
PEObs = Union[PEEvaderObs, PEPursuerObs]


class PursuitEvasionEnv(DefaultEnv):
    """The Pursuit-Evasion Grid World Environment.

    An adversarial 2D grid world problem involving two agents: an evader and a pursuer.
    The evader's goal is to reach a goal location, on the other side of the grid, while
    the goal of the pursuer is to spot the evader before it reaches it's goal. The
    evader is considered caught if it is observed by the pursuer, or occupies the same
    location. The evader and pursuer have knowledge of each others starting locations.
    However, only the evader has knowledge of it's goal location. The pursuer only
    knowns that the evader's goal location is somewhere on the opposite side of the grid
    to the evaders start location.

    This environment requires each agent to reason about the which path the other agent
    will take through the dense grid environment.

    Possible Agents
    ---------------
    - Evader = `"0"`
    - Pursuer = `"1"`

    State Space
    -----------
    Each state is made up of:

    0. the `(x, y)` coordinate of the evader
    1. the direction the evader is facing
    2. the `(x, y)` coordinate of the pursuer
    3. the direction the pursuer is facing
    4. the `(x, y)` coordinate of the evader's start location
    5. the `(x, y)` coordinate of the pursuers's start location
    6. the `(x, y)` coordinate of the evaders's goal location
    7. the minimum distance to it's goal along the shortest path achieved by the evader
       in the current episode (this is needed to correctly reward the agent for making
       progress.)

    Action Space
    ------------
    Each agent has 4 actions corresponding to moving in the 4 available directions with
    respect to the direction the agent is currently facing: `FORWARD=0`, `BACKWARDS=1`,
    `LEFT=2`, `RIGHT=3`.

    Observation Space
    -----------------
    Each agent observes:

    1. whether there is a wall (`1`) or not (`0`) in the adjacent cells in the four
       cardinal directions,
    2. whether they see the other agent in a cone in front of them (`1`) or not (`0`).
       The cone projects forward up to 'max_obs_distance' (default=`12`) cells in front
       of the agent.
    3. whether they hear the other agent (`1`) or not (`0`). The other agent is heard if
       they are within distance 2 from the agent in any direction.
    4. the `(x, y)` coordinate of the evader's start location,
    5. the `(x, y)` coordinate of the pursuer's start location,
    6. Evader: the `(x, y)` coordinate of the evader's goal location.
       Pursuer: blank coordinate `(0, 0)`.

    Note, the goal and start coordinate observations do not change during a single
    episode, but they do change between episodes.

    Rewards
    -------
    The environment is zero-sum with the pursuer receiving the negative of the evader
    reward. Additionally, rewards are by default normalized so that returns are bounded
    between `-1` and `1` (this can be disabled by the `normalize_reward` parameter).

    The evader receives a reward of `1` for reaching it's goal location and a
    reward of `-1` if it gets captured. Additionally, the evader receives a small
    reward of `0.01` each time it's minimum distance achieved to it's goal along the
    shortest path decreases for the current episode. This is to make it so the
    environment is no longer sparesely rewarded and helps with exploration and learning
    (it can be disabled by the `use_progress_reward` parameter.)

    Dynamics
    --------
    By default actions are deterministic and will move the agent one cell in the target
    direction if the cell is empty.

    The environment can also be run in stochastic mode by changing the `action_probs`
    parameter at initialization. This controls the probability the agent will move in
    the desired direction each step, otherwise moving randomly in one of the other 3
    possible directions.

    Starting State
    --------------
    At the start of each episode the start location of the evader is selected at random
    from all possible start locations. The evader's goal location is then chosen
    randomly from the set of available goal locations given the evaders start location
    (in the default maps the goal location is always on the opposite side of the map
    from a start location). The pursuers start location is similarly chosen from it's
    set of possible start locations.

    Episode End
    -----------
    An episode ends when either the evader is caught or the evader reaches it's goal.
    By default a `max_episode_steps` limit of `100` steps is also set. This may need to
    be adjusted when using larger grids (this can be done by manually specifying a value
    for `max_episode_steps` when creating the environment with `posggym.make`).

    Arguments
    ---------

    - `grid` - the grid layout to use. This can either be a string specifying one of
         the supported grids, or a custom :class:`PEGrid` object (default = `"16x16"`).
    - `action_probs` - the action success probability for each agent. This can be a
        single float (same value for both evader and pursuer agents) or a tuple with
        separate values for each agent (default = `1.0`).
    - `max_obs_distance` - the maximum number of cells in front each agent's field of
        vision extends (default = `12`).
    - `num_prey` - the number of prey (default = `3`)
    - `normalize_reward` - whether to normalize both agents' rewards to be between `-1`
        and `1` (default = 'True`)
    - `use_progress_reward` - whether to reward the evader agent for making progress
        towards it's goal. If False the evader will only be rewarded when it reaches
        it's goal, making it a sparse reward problem (default = 'True`).

    Available variants
    ------------------

    The PursuitEvasion environment comes with a number of pre-built grid layouts which
    can be passed as an argument to `posggym.make`, to create different grids.

    | Grid name         | Grid size |
    |-------------------|-----------|
    | `5x5`             | 8x8       |
    | `16x16`           | 16x16     |
    | `32x32`           | 32x32     |

    For example to use the PursuitEvasion environment with the `32x32` grid layout, and
    episode step limit of 200, and the default values for the other parameters you would
    use:

    ```python
    import posggym
    env = posggym.make(
        'PursuitEvasion-v0',
        max_episode_steps=200,
        grid="32x32",
    )
    ```

    References
    ----------
    - [This Pursuit-Evasion implementation is directly inspired by the problem] Seaman,
      Iris Rubi, Jan-Willem van de Meent, and David Wingate. 2018. “Nested Reasoning
      About Autonomous Agents Using Probabilistic Programs.”
      ArXiv Preprint ArXiv:1812.01569.
    - Schwartz, Jonathon, Ruijia Zhou, and Hanna Kurniawati. "Online Planning for
      Interactive-POMDPs using Nested Monte Carlo Tree Search." In 2022 IEEE/RSJ
      International Conference on Intelligent Robots and Systems (IROS), pp. 8770-8777.
      IEEE, 2022.

    """

    metadata = {
        "render_modes": ["human", "ansi", "rgb_array"],
        "render_fps": 15,
    }

    def __init__(
        self,
        grid: Union[str, "PEGrid"] = "16x16",
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        max_obs_distance: int = 12,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
        render_mode: Optional[str] = None,
        **kwargs,
    ):
        model = PursuitEvasionModel(
            grid,
            action_probs=action_probs,
            max_obs_distance=max_obs_distance,
            normalize_reward=normalize_reward,
            use_progress_reward=use_progress_reward,
            **kwargs,
        )
        super().__init__(model, render_mode=render_mode)

        self.max_obs_distance = max_obs_distance
        fov_width = model.grid.get_max_fov_width(
            model.FOV_EXPANSION_INCR, max_obs_distance
        )
        self._obs_dims = (
            min(max_obs_distance, max(model.grid.width, model.grid.height)),
            0,
            fov_width // 2,
            fov_width // 2,
        )
        self.renderer = None
        self._agent_imgs = None

    def reset(
        self, *, seed: Optional[int] = None, options: Optional[Dict[str, Any]] = None
    ) -> Tuple[Dict[str, M.ObsType], Dict[str, Dict]]:
        # reset renderer since goal location can change between episodes
        self._renderer = None
        return super().reset(seed=seed, options=options)

    def render(self):
        if self.render_mode is None:
            assert self.spec is not None
            logger.warn(
                "You are calling render method without specifying any render mode. "
                "You can specify the render_mode at initialization, "
                f'e.g. posggym.make("{self.spec.id}", render_mode="rgb_array")'
            )
            return
        if self.render_mode == "ansi":
            return self._render_ansi()
        return self._render_img()

    def _render_ansi(self):
        evader_coord = self._state[0]
        pursuer_coord = self._state[2]
        goal_coord = self._state[6]
        grid = self.model.grid  # type: ignore

        grid_str = grid.get_ascii_repr(goal_coord, evader_coord, pursuer_coord)
        output = [
            f"Step: {self._step_num}",
            grid_str,
        ]
        if self._last_actions is not None:
            action_str = ", ".join(
                [str(Direction(a)) for a in self._last_actions.values()]
            )
            output.insert(1, f"Actions: <{action_str}>")
            output.append(f"Rewards: <{self._last_rewards}>")
        return "\n".join(output) + "\n"

    def _render_img(self):
        evader_coord = self._state[0]
        pursuer_coord = self._state[2]
        goal_coord = self._state[6]
        model: PursuitEvasionModel = self.model  # type: ignore

        import posggym.envs.grid_world.render as render_lib

        if self.renderer is None:
            self.renderer = render_lib.GWRenderer(
                self.render_mode,
                model.grid,
                render_fps=self.metadata["render_fps"],
                env_name="PursuitEvasion",
            )

        if self._agent_imgs is None:
            self._agent_imgs = {
                i: render_lib.GWTriangle(
                    (0, 0),
                    self.renderer.cell_size,
                    render_lib.get_agent_color(i)[0],
                    Direction.NORTH,
                )
                for i in self.possible_agents
            }

        observed_coords = []
        for i in range(len(self.agents)):
            observed_coords.extend(
                model.grid.get_fov(
                    self._state[2 * i],
                    self._state[2 * i + 1],
                    self.model.FOV_EXPANSION_INCR,
                    self.max_obs_distance,
                )
            )

        render_objects = [
            render_lib.GWRectangle(
                goal_coord, self.renderer.cell_size, render_lib.get_color("green")
            )
        ]

        agent_coords_and_dirs = {
            "0": (evader_coord, self._state[1]),
            "1": (pursuer_coord, self._state[3]),
        }
        for i, (c, d) in agent_coords_and_dirs.items():
            agent_img = self._agent_imgs[i]
            agent_img.coord = c
            agent_img.facing_dir = d
            render_objects.append(agent_img)

        return self.renderer.render(render_objects, observed_coords)

    def close(self) -> None:
        if self.renderer is not None:
            self.renderer.close()
            self.renderer = None


class PursuitEvasionModel(M.POSGModel[PEState, PEObs, PEAction]):
    """Discrete Pursuit-Evasion Model."""

    NUM_AGENTS = 2

    EVADER_IDX = 0
    PURSUER_IDX = 1

    # Evader-centric rewards
    R_PROGRESS = 0.01  # Reward making progress toward goal
    R_EVADER_ACTION = -0.01  # Reward each step for evader
    R_PURSUER_ACTION = -0.01  # Reward each step for pursuer
    R_CAPTURE = -1.0  # Reward for being captured
    R_EVASION = 1.0  # Reward for reaching goal

    FOV_EXPANSION_INCR = 3
    HEARING_DIST = 2

    def __init__(
        self,
        grid: Union[str, "PEGrid"],
        action_probs: Union[float, Tuple[float, float]] = 1.0,
        max_obs_distance: int = 12,
        normalize_reward: bool = True,
        use_progress_reward: bool = True,
    ):
        if isinstance(grid, str):
            assert grid in SUPPORTED_GRIDS, (
                f"Unsupported grid name '{grid}'. If grid is a string it must be one "
                f"of: {SUPPORTED_GRIDS.keys()}."
            )
            grid = SUPPORTED_GRIDS[grid][0]()

        if isinstance(action_probs, float):
            action_probs = (action_probs, action_probs)

        self._grid = grid
        self.action_probs = action_probs
        self.max_obs_distance = max_obs_distance
        self.normalize_reward = normalize_reward
        self.use_progress_reward = use_progress_reward

        self._max_sp_distance = self._grid.get_max_shortest_path_distance()
        self._max_raw_return = self.R_EVASION
        if self.use_progress_reward:
            self._max_raw_return += (self._max_sp_distance + 1) * self.R_PROGRESS
        self._min_raw_return = -self._max_raw_return

        def _coord_space():
            return spaces.Tuple(
                (spaces.Discrete(self.grid.width), spaces.Discrete(self.grid.height))
            )

        self.possible_agents = tuple(str(i) for i in range(self.NUM_AGENTS))
        # s = Tuple[Coord, Direction, Coord, Direction, Coord, Coord, Coord, int]
        # e_coord, e_dir, p_coord, p_dir, e_start, p_start, e_goal, max_sp
        self.state_space = spaces.Tuple(
            (
                _coord_space(),
                spaces.Discrete(len(Direction)),
                _coord_space(),
                spaces.Discrete(len(Direction)),
                _coord_space(),
                _coord_space(),
                _coord_space(),
                spaces.Discrete(self._max_sp_distance + 1),
            )
        )
        self.action_spaces = {
            i: spaces.Discrete(len(Direction)) for i in self.possible_agents
        }
        # o = Tuple[Tuple[WallObs, seen , heard], Coord, Coord, Coord]
        # Wall obs, seen, heard, e_start, p_start, e_goal/blank
        self.observation_spaces = {
            i: spaces.Tuple(
                (
                    spaces.Tuple(tuple(spaces.Discrete(2) for _ in range(6))),
                    _coord_space(),
                    _coord_space(),
                    _coord_space(),
                )
            )
            for i in self.possible_agents
        }
        self.is_symmetric = False

    @property
    def grid(self) -> "PEGrid":
        """The underlying grid for this model instance."""
        return self._grid

    @property
    def reward_ranges(self) -> Dict[str, Tuple[float, float]]:
        max_reward = self.R_EVASION
        if self.use_progress_reward:
            max_reward += self.R_PROGRESS
        if self.normalize_reward:
            max_reward = self._get_normalized_reward(max_reward)
        return {i: (-max_reward, max_reward) for i in self.possible_agents}

    @property
    def rng(self) -> seeding.RNG:
        if self._rng is None:
            self._rng, seed = seeding.std_random()
        return self._rng

    def get_agents(self, state: PEState) -> List[str]:
        return list(self.possible_agents)

    def sample_initial_state(self) -> PEState:
        return self._sample_initial_state(None, None, None)

    def sample_agent_initial_state(self, agent_id: str, obs: PEObs) -> PEState:
        if agent_id == self.EVADER_IDX:
            return self._sample_initial_state(
                evader_coord=obs[1], pursuer_coord=obs[2], goal_coord=obs[3]
            )
        return self._sample_initial_state(
            evader_coord=obs[1], pursuer_coord=obs[2], goal_coord=None
        )

    def _sample_initial_state(
        self,
        evader_coord: Optional[Coord],
        pursuer_coord: Optional[Coord],
        goal_coord: Optional[Coord],
    ) -> PEState:
        if evader_coord is None:
            evader_coord = self.rng.choice(self.grid.evader_start_coords)
        if pursuer_coord is None:
            pursuer_coord = self.rng.choice(self.grid.pursuer_start_coords)
        if goal_coord is None:
            goal_coord = self.rng.choice(self.grid.get_goal_coords(evader_coord))
        return PEState(
            evader_coord,
            INITIAL_DIR,
            pursuer_coord,
            INITIAL_DIR,
            evader_coord,
            pursuer_coord,
            goal_coord,
            self.grid.get_shortest_path_distance(evader_coord, goal_coord),
        )

    def sample_initial_obs(self, state: PEState) -> Dict[str, PEObs]:
        return self._get_obs(state)[0]

    def step(
        self, state: PEState, actions: Dict[str, PEAction]
    ) -> M.JointTimestep[PEState, PEObs]:
        assert all(0 <= a_i < len(Direction) for a_i in actions.values())
        next_state = self._get_next_state(state, actions)
        obs, evader_detected = self._get_obs(next_state)
        rewards = self._get_reward(state, next_state, evader_detected)
        all_done = self._is_done(next_state)
        terminated = {i: all_done for i in self.possible_agents}
        truncated = {i: False for i in self.possible_agents}
        info: Dict[str, Dict] = {i: {} for i in self.possible_agents}
        if all_done:
            for i, outcome in self._get_outcome(next_state).items():
                info[i]["outcome"] = outcome
        return M.JointTimestep(
            next_state, obs, rewards, terminated, truncated, all_done, info
        )

    def _get_next_state(self, state: PEState, actions: Dict[str, PEAction]) -> PEState:
        evader_a = actions[str(self.EVADER_IDX)]
        pursuer_a = actions[str(self.PURSUER_IDX)]

        if (
            self.action_probs[self.EVADER_IDX] < 1.0
            and self.rng.random() > self.action_probs[self.EVADER_IDX]
        ):
            other_as = [a for a in range(len(Direction)) if a != evader_a]
            evader_a = self.rng.choice(other_as)

        if (
            self.action_probs[self.PURSUER_IDX]
            and self.rng.random() > self.action_probs[self.PURSUER_IDX]
        ):
            other_as = [a for a in range(len(Direction)) if a != pursuer_a]
            pursuer_a = self.rng.choice(other_as)

        pursuer_next_dir = Direction(ACTION_TO_DIR[pursuer_a][state.pursuer_dir])
        pursuer_next_coord = self.grid.get_next_coord(
            state.pursuer_coord, pursuer_next_dir, ignore_blocks=False
        )

        evader_next_coord = state.evader_coord
        evader_next_dir = Direction(ACTION_TO_DIR[evader_a][state.evader_dir])
        if pursuer_next_coord != state.evader_coord:
            evader_next_coord = self.grid.get_next_coord(
                state.evader_coord, evader_next_dir, ignore_blocks=False
            )

        min_sp_distance = min(
            state.min_goal_dist,
            self.grid.get_shortest_path_distance(
                evader_next_coord, state.evader_goal_coord
            ),
        )

        return PEState(
            evader_next_coord,
            evader_next_dir,
            pursuer_next_coord,
            pursuer_next_dir,
            state.evader_start_coord,
            state.pursuer_start_coord,
            state.evader_goal_coord,
            min_sp_distance,
        )

    def _get_obs(self, state: PEState) -> Tuple[Dict[str, PEObs], bool]:
        walls, seen, heard = self._get_agent_obs(
            state.evader_coord, state.evader_dir, state.pursuer_coord
        )
        evader_obs = (
            (*walls, seen, heard),
            state.evader_start_coord,
            state.pursuer_start_coord,
            state.evader_goal_coord,
        )

        walls, seen, heard = self._get_agent_obs(
            state.pursuer_coord, state.pursuer_dir, state.evader_coord
        )
        pursuer_obs = (
            (*walls, seen, heard),
            state.evader_start_coord,
            state.pursuer_start_coord,
            # pursuer doesn't observe evader goal coord
            (0, 0),
        )
        return {
            str(self.EVADER_IDX): evader_obs,
            str(self.PURSUER_IDX): pursuer_obs,
        }, bool(seen)

    def _get_agent_obs(
        self, agent_coord: Coord, agent_dir: Direction, opp_coord: Coord
    ) -> Tuple[Tuple[int, int, int, int], int, int]:
        adj_coords = self.grid.get_neighbours(
            agent_coord, ignore_blocks=True, include_out_of_bounds=True
        )
        walls: Tuple[int, int, int, int] = tuple(  # type: ignore
            int(not self.grid.coord_in_bounds(coord) or coord in self.grid.block_coords)
            for coord in adj_coords
        )
        seen = self._get_opponent_seen(agent_coord, agent_dir, opp_coord)
        dist = self.grid.manhattan_dist(agent_coord, opp_coord)
        heard = dist <= self.HEARING_DIST
        return walls, int(seen), int(heard)

    def _get_opponent_seen(
        self, ego_coord: Coord, ego_dir: Direction, opp_coord: Coord
    ) -> bool:
        fov = self.grid.get_fov(
            ego_coord, ego_dir, self.FOV_EXPANSION_INCR, self.max_obs_distance
        )
        return opp_coord in fov

    def _get_reward(
        self, state: PEState, next_state: PEState, evader_seen: bool
    ) -> Dict[str, float]:
        evader_coord = next_state.evader_coord
        pursuer_coord = next_state.pursuer_coord
        evader_goal_coord = next_state.evader_goal_coord

        evader_reward = 0.0
        if self.use_progress_reward and next_state.min_goal_dist < state.min_goal_dist:
            evader_reward += self.R_PROGRESS

        if evader_coord == pursuer_coord or evader_seen:
            evader_reward += self.R_CAPTURE
        elif evader_coord == evader_goal_coord:
            evader_reward += self.R_EVASION

        if self.normalize_reward:
            evader_reward = self._get_normalized_reward(evader_reward)

        return {
            str(self.EVADER_IDX): evader_reward,
            str(self.PURSUER_IDX): -evader_reward,
        }

    def _is_done(self, state: PEState) -> bool:
        evader_coord, pursuer_coord = state.evader_coord, state.pursuer_coord
        pursuer_dir = state.pursuer_dir
        return (
            evader_coord == pursuer_coord
            or evader_coord == state.evader_goal_coord
            or self._get_opponent_seen(pursuer_coord, pursuer_dir, evader_coord)
        )

    def _get_outcome(self, state: PEState) -> Dict[str, M.Outcome]:
        # Assuming this method is called on final timestep
        evader_coord, pursuer_coord = state.evader_coord, state.pursuer_coord
        evader_goal_coord = state.evader_goal_coord
        pursuer_dir = state.pursuer_dir
        evader_id, pursuer_id = str(self.EVADER_IDX), str(self.PURSUER_IDX)
        # check this first before relatively expensive detection check
        if evader_coord == pursuer_coord:
            return {evader_id: M.Outcome.LOSS, pursuer_id: M.Outcome.WIN}
        if evader_coord == evader_goal_coord:
            return {evader_id: M.Outcome.WIN, pursuer_id: M.Outcome.LOSS}
        if self._get_opponent_seen(pursuer_coord, pursuer_dir, evader_coord):
            return {evader_id: M.Outcome.LOSS, pursuer_id: M.Outcome.WIN}
        return {evader_id: M.Outcome.DRAW, pursuer_id: M.Outcome.DRAW}

    def _get_normalized_reward(self, reward: float) -> float:
        """Normalize reward in [-1, 1] interval."""
        diff = self._max_raw_return - self._min_raw_return
        return 2 * (reward - self._min_raw_return) / diff - 1


class PEGrid(Grid):
    """A grid for the Pursuit Evasion Problem."""

    def __init__(
        self,
        grid_width: int,
        grid_height: int,
        block_coords: Set[Coord],
        goal_coords_map: Dict[Coord, List[Coord]],
        evader_start_coords: List[Coord],
        pursuer_start_coords: List[Coord],
    ):
        super().__init__(grid_width, grid_height, block_coords)
        self._goal_coords_map = goal_coords_map
        self.evader_start_coords = evader_start_coords
        self.pursuer_start_coords = pursuer_start_coords
        self.shortest_paths = self.get_all_shortest_paths(self.all_goal_coords)

    @property
    def all_goal_coords(self) -> List[Coord]:
        """The list of all evader goal locations."""
        all_locs = set()
        for v in self._goal_coords_map.values():
            all_locs.update(v)
        return list(all_locs)

    def get_goal_coords(self, evader_start_coord: Coord) -> List[Coord]:
        """Get list of possible evader goal coords for given start coords."""
        return self._goal_coords_map[evader_start_coord]

    def get_shortest_path_distance(self, coord: Coord, dest: Coord) -> int:
        """Get the shortest path distance from coord to destination."""
        return int(self.shortest_paths[dest][coord])

    def get_max_shortest_path_distance(self) -> int:
        """Get max shortest path distance between any start and goal coords."""
        max_dist = 0
        for start_coord, dest_coords in self._goal_coords_map.items():
            max_dist = max(
                max_dist,
                int(
                    max(
                        [
                            self.get_shortest_path_distance(start_coord, dest)
                            for dest in dest_coords
                        ]
                    )
                ),
            )
        return max_dist

    def get_ascii_repr(
        self,
        goal_coord: Union[None, Coord, List[Coord]],
        evader_coord: Union[None, Coord, List[Coord]],
        pursuer_coord: Union[None, Coord, List[Coord]],
    ) -> str:
        """Get ascii repr of grid."""
        if goal_coord is None:
            goal_coords = set()
        elif isinstance(goal_coord, list):
            goal_coords = set(goal_coord)
        else:
            goal_coords = {goal_coord}

        grid_repr = []
        for row in range(self.height):
            row_repr = []
            for col in range(self.width):
                coord = (col, row)
                if coord in goal_coords:
                    row_repr.append("G")
                elif coord in self.block_coords:
                    row_repr.append("#")
                else:
                    row_repr.append(".")
            grid_repr.append(row_repr)

        if evader_coord is None:
            evader_coord = []
        elif not isinstance(evader_coord, List):
            evader_coord = [evader_coord]

        for coord in evader_coord:
            grid_repr[coord[1]][coord[0]] = "R"

        if pursuer_coord is None:
            pursuer_coord = []
        elif not isinstance(pursuer_coord, List):
            pursuer_coord = [pursuer_coord]

        for coord in pursuer_coord:
            if coord in evader_coord:
                grid_repr[coord[1]][coord[0]] = "X"
            else:
                grid_repr[coord[1]][coord[0]] = "C"

        return "\n".join([" ".join(r) for r in grid_repr])

    def get_init_ascii_repr(self) -> str:
        """Get ascii repr of initial grid."""
        return self.get_ascii_repr(
            self.all_goal_coords, self.evader_start_coords, self.pursuer_start_coords
        )

    def get_fov(
        self,
        origin: Coord,
        direction: Direction,
        widening_increment: int,
        max_depth: int,
    ) -> Set[Coord]:
        """Get the Field of vision from origin looking in given direction.

        Uses BFS starting from origin and expanding in the direction, while
        accounting for obstacles blocking the field of view,
        to get the field of vision.
        """
        assert widening_increment > 0
        assert max_depth > 0
        fov = {origin}

        frontier_queue: Deque[Coord] = deque([origin])
        visited = {origin}

        while len(frontier_queue):
            coord = frontier_queue.pop()

            for next_coord in self._get_fov_successors(
                origin, direction, coord, widening_increment, max_depth
            ):
                if next_coord not in visited:
                    visited.add(next_coord)
                    frontier_queue.append(next_coord)
                    fov.add(next_coord)
        return fov

    def _get_fov_successors(
        self,
        origin: Coord,
        direction: Direction,
        coord: Coord,
        widening_increment: int,
        max_depth: int,
    ) -> List[Coord]:
        if direction in [Direction.NORTH, Direction.SOUTH]:
            depth = abs(origin[1] - coord[1])
        else:
            depth = abs(origin[0] - coord[0])

        if depth >= max_depth:
            return []

        successors = []
        forward_successor = self._get_fov_successor(coord, direction)
        if forward_successor is None:
            return []
        else:
            successors.append(forward_successor)

        # Expands FOV at depth 1 and then every widening_increment depth
        if depth != 1 and depth % widening_increment != 0:
            # Don't expand sideways
            return successors

        side_coords_list: List[Coord] = []

        if direction in [Direction.NORTH, Direction.SOUTH]:
            if 0 < coord[0] <= origin[0]:
                side_coords_list.append((coord[0] - 1, coord[1]))
            if origin[0] <= coord[0] < self.width - 1:
                side_coords_list.append((coord[0] + 1, coord[1]))
        else:
            if 0 < coord[1] <= origin[1]:
                side_coords_list.append((coord[0], coord[1] - 1))
            elif origin[1] <= coord[1] < self.height - 1:
                side_coords_list.append((coord[0], coord[1] + 1))

        side_successor: Optional[Coord] = None
        for side_coord in side_coords_list:
            if side_coord in self.block_coords:
                continue

            side_successor = self._get_fov_successor(side_coord, direction)
            if side_successor is not None:
                successors.append(side_successor)

        return successors

    def _get_fov_successor(self, coord: Coord, direction: Direction) -> Optional[Coord]:
        new_coord = self.get_next_coord(coord, direction, ignore_blocks=False)
        if new_coord == coord:
            # move in given direction is blocked or out-of-bounds
            return None
        return new_coord

    def get_max_fov_width(self, widening_increment: int, max_depth: int) -> int:
        """Get the maximum width of field of vision."""
        max_width = 1
        for d in range(max_depth):
            if d == 1 or d % widening_increment == 0:
                max_width += 2
        return min(max_depth, min(self.width, self.height))


def get_8x8_grid() -> PEGrid:
    """Generate the 8-by-8 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.
    """
    ascii_map = (
        "  9 #8 #"
        "# #    #"
        "# ## # 7"
        "  6   # "
        "# #5# # "
        "   #    "
        "#    #2 "
        "0 #1### "
    )

    return _convert_map_to_grid(ascii_map, 8, 8)


def get_16x16_grid() -> PEGrid:
    """Generate the 16-by-16 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.

    """
    ascii_map = (
        "  ## ####### ###"
        "    9##    8 ###"
        "##       # # ###"
        "## ## ##      ##"
        "## ## ##   ## ##"
        "#  #  ##   ## 7#"
        "# ## ###   #   #"
        "  #  6       # #"
        "#   ## ##  ##  #"
        "##   #5##  #   #"
        "## #   #     # #"
        "   # #   ##    #"
        " # # ##  ##   ##"
        "0#       #     #"
        "   ### #   ##2  "
        "###    1 #####  "
    )
    return _convert_map_to_grid(ascii_map, 16, 16)


def get_32x32_grid() -> PEGrid:
    """Generate the 32-by-32 PE grid layout.

    This is an approximate discrete version of the map layout from:
    Seaman et al 2018, 'Nested Reasoning About Autonomous Agents Using
    Probabilistic Programs'

    - 0, 1, 2, 7, 8, 9 are possible evader start and goal locations
    - 5, 6 are possible pursuer start locations

    The evader start and goal locations are always on opposite sides of the
    map.

    """
    ascii_map = (
        "       #  ########### #         "
        "   #####  ########### #  #######"
        "      #   ######      8  #######"
        " #       9#               ######"
        " ##              #### ##  ######"
        " ##   #####  ##  #### ##   #####"
        " ##   #####  ##  ##        #####"
        "      ##### ####      ###  #####"
        " ### ###### ####      ###   ####"
        " ### #####  ####      ####  ####"
        " ##  ##### #####      ##### 7###"
        " ##  ####  #####       ####  ###"
        " #  #####  #####       ####  ###"
        " #  ##### ######       #     ###"
        " #       6               ##  ###"
        "    #       #   ##      ####  ##"
        "     #   ####  ###     ####   ##"
        "#### #   ####  ###    ####    ##"
        "#### #   ### 5####   ####     ##"
        "####  #      ####     ##  #   ##"
        "##### #      ####        ##   ##"
        "#                          ##  #"
        "         ###        ##    2 #  #"
        "  ###### ###      #### ## #    #"
        "########  ##      ####    #  #  "
        "########  ##       ####     ####"
        "###          ##   1##        ###"
        "          #  ####       #       "
        "   0  ###### ######   ####      "
        "  ########## #        #####     "
        "##########      ############    "
        "#                               "
    )
    return _convert_map_to_grid(ascii_map, 32, 32)


def _loc_to_coord(loc: int, grid_width: int) -> Coord:
    return (loc % grid_width, loc // grid_width)


def _convert_map_to_grid(
    ascii_map: str,
    height: int,
    width: int,
    block_symbol: str = "#",
    pursuer_start_symbols: Optional[Set[str]] = None,
    evader_start_symbols: Optional[Set[str]] = None,
    evader_goal_symbol_map: Optional[Dict] = None,
) -> PEGrid:
    assert len(ascii_map) == height * width

    if pursuer_start_symbols is None:
        pursuer_start_symbols = {"3", "4", "5", "6"}
    if evader_start_symbols is None:
        evader_start_symbols = {"0", "1", "2", "7", "8", "9"}
    if evader_goal_symbol_map is None:
        evader_goal_symbol_map = {
            "0": ["7", "8", "9"],
            "1": ["7", "8", "9"],
            "2": ["8", "9"],
            "7": ["0", "1"],
            "8": ["0", "1", "2"],
            "9": ["0", "1", "2"],
        }

    block_coords = set()
    evader_start_coords = []
    pursuer_start_coords = []
    evader_symbol_coord_map = {}

    for loc, symbol in enumerate(ascii_map):
        coord = _loc_to_coord(loc, width)
        if symbol == block_symbol:
            block_coords.add(coord)
        elif symbol in pursuer_start_symbols:
            pursuer_start_coords.append(coord)
        elif symbol in evader_start_symbols:
            evader_start_coords.append(coord)
            evader_symbol_coord_map[symbol] = coord

    evader_goal_coords_map = {}
    for start_symbol, goal_symbols in evader_goal_symbol_map.items():
        start_coord = evader_symbol_coord_map[start_symbol]
        evader_goal_coords_map[start_coord] = [
            evader_symbol_coord_map[symbol] for symbol in goal_symbols
        ]

    return PEGrid(
        grid_width=width,
        grid_height=height,
        block_coords=block_coords,
        goal_coords_map=evader_goal_coords_map,
        evader_start_coords=evader_start_coords,
        pursuer_start_coords=pursuer_start_coords,
    )


# grid_name: (grid_make_fn, step_limit)
SUPPORTED_GRIDS = {
    "8x8": (get_8x8_grid, 50),
    "16x16": (get_16x16_grid, 100),
    "32x32": (get_32x32_grid, 200),
}



"""Module of wrapper classes.

Ref:
https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/__init__.py
"""
from posggym.wrappers.discretize_actions import DiscretizeActions
from posggym.wrappers.env_checker import PassiveEnvChecker
from posggym.wrappers.flatten_observations import FlattenObservations
from posggym.wrappers.order_enforcing import OrderEnforcing
from posggym.wrappers.record_episode_statistics import RecordEpisodeStatistics
from posggym.wrappers.record_video import RecordVideo
from posggym.wrappers.rescale_actions import RescaleActions
from posggym.wrappers.rescale_observations import RescaleObservations
from posggym.wrappers.time_limit import TimeLimit



"""Wrapper for rescaling observations to within min and max values."""
from typing import Dict, Union

import numpy as np
from gymnasium import spaces

from posggym import Env, ObservationWrapper


class RescaleObservations(ObservationWrapper):
    """Observation wrapper that rescales observations to the range [min_obs, max_obs].

    The base environment :attr:`env` must have an observation space of type
    :class:`spaces.Box` for each agent. If :attr:`min_obs` or :attr:`max_obs` are numpy
    arrays, the shape must match the shape of the environment's observation space for
    the given agent. If :attr:`min_obs` or :attr:`max_obs` are dictionaries then they
    must have an entry for each possible agent ID in the wrapped environment.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper
    min_obs : float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        The minimum value for the scaled observations.
    max_obs : float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        The maximum value for the scaled observations.


    Note
    ----
    Explanation of how to scale number from one interval into new interval:
    https://stats.stackexchange.com/questions/281162/scale-a-number-between-a-range

    """

    def __init__(
        self,
        env: Env,
        min_obs: Union[
            float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        ],
        max_obs: Union[
            float, int, np.ndarray, Dict[str, Union[float, int, np.ndarray]]
        ],
    ):
        self.min_obs = {}
        self.max_obs = {}
        self.rescale_factor = {}
        for i, obs_space in env.observation_spaces.items():
            assert isinstance(obs_space, spaces.Box), (
                f"Expected Box observation space for all agents, got {type(obs_space)} "
                f"for agent '{i}'."
            )

            min_obs_i = min_obs[i] if isinstance(min_obs, dict) else min_obs
            if isinstance(min_obs_i, (float, int)):
                self.min_obs[i] = np.full_like(obs_space.low, min_obs_i)
            else:
                assert isinstance(min_obs_i, np.ndarray), min_obs_i
                assert min_obs_i.shape == obs_space.shape, (
                    min_obs_i.shape,
                    obs_space.shape,
                )
                self.min_obs[i] = min_obs_i

            max_obs_i = max_obs[i] if isinstance(max_obs, dict) else max_obs
            if isinstance(max_obs_i, (float, int)):
                self.max_obs[i] = np.full_like(obs_space.high, max_obs_i)
            else:
                assert isinstance(max_obs_i, np.ndarray), max_obs_i
                assert max_obs_i.shape == obs_space.shape, (
                    max_obs_i.shape,
                    obs_space.shape,
                )
                self.max_obs[i] = max_obs_i

            assert np.less_equal(self.min_obs[i], self.max_obs[i]).all(), (
                self.min_obs[i],
                self.max_obs[i],
            )
            self.rescale_factor[i] = (self.max_obs[i] - self.min_obs[i]) / (
                obs_space.high - obs_space.low
            )

        super().__init__(env)
        self._observation_spaces = {
            i: spaces.Box(low=self.min_obs[i], high=self.max_obs[i])
            for i in env.observation_spaces
        }

    def observations(self, obs):
        rescaled_obs = {}
        for i, obs_i in obs.items():
            low = self.env.observation_spaces[i].low
            rescaled_obs[i] = (obs_i - low) * self.rescale_factor[i] + self.min_obs[i]
        return rescaled_obs



"""Shortest path policy for PursuitEvasion env."""
from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, cast

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.grid_world.pursuit_evasion import (
    ACTION_TO_DIR,
    INITIAL_DIR,
    PEAction,
    PEObs,
    PursuitEvasionModel,
)
from posggym.utils import seeding

if TYPE_CHECKING:
    from posggym.envs.grid_world.core import Coord, Direction
    from posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


class PEShortestPathPolicy(Policy[PEAction, PEObs]):
    """Shortest Path Policy for the Pursuit Evasion environment.

    This policy sets the preferred action as the one which is on the shortest
    path to the evaders goal and which doesn't leave agent in same position.

    """

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID, **kwargs):
        super().__init__(model, agent_id, policy_id)
        self.model = cast(PursuitEvasionModel, model)
        assert all(p == 1.0 for p in self.model.action_probs), (
            f"{self.__class__.__name__} only supported for deterministic versions of "
            "the PursuitEvasion environment."
        )

        self._grid = self.model.grid
        self._is_evader = self.agent_id == self.model.EVADER_IDX

        self._action_space = list(
            range(self.model.action_spaces[self.agent_id].n)  # type: ignore
        )

        evader_end_coords = list(
            set(self._grid.evader_start_coords + self._grid.all_goal_coords)
        )
        self._dists = self._grid.get_all_shortest_paths(evader_end_coords)

        self._rng, _ = seeding.std_random()

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.std_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update(
            {
                "action": None,
                "update_num": 0,
                "facing_dir": INITIAL_DIR,
                "coord": (0, 0),
                "prev_coord": (0, 0),
                "target_coord": (0, 0),
            }
        )
        return state

    def get_next_state(
        self,
        obs: PEObs,
        state: PolicyState,
    ) -> PolicyState:
        if state["update_num"] == 0:
            # I.e. first observation with no action performed yet
            next_update_num = 1
            next_facing_dir = INITIAL_DIR
            next_coord = obs[1] if self._is_evader else obs[2]
            next_prev_coord = (0, 0)
            target_coord = obs[3] if self._is_evader else obs[1]
        else:
            next_update_num = state["update_num"] + 1
            next_facing_dir = ACTION_TO_DIR[state["action"]][state["facing_dir"]]
            next_coord = self._grid.get_next_coord(
                state["coord"], next_facing_dir, False
            )
            next_prev_coord = state["coord"]
            target_coord = state["target_coord"]

        next_action = self._get_sp_action(
            next_facing_dir,
            next_coord,
            next_prev_coord,
            target_coord,
        )
        next_state = {
            "update_num": next_update_num,
            "facing_dir": next_facing_dir,
            "coord": next_coord,
            "prev_coord": next_prev_coord,
            "target_coord": target_coord,
            "action": next_action,
        }
        return next_state

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        last_action, obs = history.get_last_step()
        assert obs is not None
        facing_dir = INITIAL_DIR
        coord = obs[1] if self._is_evader else obs[2]
        prev_coord = coord
        target_coord = obs[3] if self._is_evader else obs[1]
        for a, _ in history:
            if a is None:
                continue
            facing_dir = ACTION_TO_DIR[a][facing_dir]
            prev_coord = coord
            coord = self._grid.get_next_coord(coord, facing_dir, False)
        return {
            "action": last_action,
            "update_num": len(history),
            "facing_dir": facing_dir,
            "coord": coord,
            "prev_coord": prev_coord,
            "target_coord": target_coord,
        }

    def sample_action(self, state: PolicyState) -> PEAction:
        return state["action"]

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        dist = self._get_pi_from_coords(
            state["facing_dir"],
            state["coord"],
            state["prev_coord"],
            state["target_coord"],
        )
        return action_distributions.DiscreteActionDistribution(dist, self._rng)

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def _get_pi_from_coords(
        self, facing_dir: Direction, coord: Coord, prev_coord: Coord, goal_coord: Coord
    ) -> Dict[PEAction, float]:
        dists = self._get_action_sp_dists(facing_dir, coord, prev_coord, goal_coord)

        pi = {a: 1.0 for a in self._action_space}
        max_dist, min_dist = max(dists), min(dists)
        dist_gap = max(1, max_dist - min_dist)

        weight_sum = 0.0
        for i, a in enumerate(self._action_space):
            weight = 1 - ((dists[i] - min_dist) / dist_gap)
            pi[a] = weight
            weight_sum += weight

        for a in self._action_space:
            pi[a] /= weight_sum

        return pi

    def _get_sp_action(
        self, facing_dir: Direction, coord: Coord, prev_coord: Coord, goal_coord: Coord
    ) -> PEAction:
        sp_action = self._action_space[0]
        sp_dist = float("inf")
        for a in self._action_space:
            a_dir = ACTION_TO_DIR[a][facing_dir]
            next_coord = self._grid.get_next_coord(coord, a_dir, False)
            if next_coord == prev_coord:
                continue

            dist = self._dists[goal_coord][next_coord]
            if dist < sp_dist:
                sp_dist = dist
                sp_action = a
        return sp_action

    def _get_action_sp_dists(
        self, facing_dir: Direction, coord: Coord, prev_coord: Coord, goal_coord: Coord
    ) -> List[int]:
        dists = []
        for a in self._action_space:
            a_dir = ACTION_TO_DIR[a][facing_dir]
            new_coord = self._grid.get_next_coord(coord, a_dir, False)
            if new_coord == prev_coord:
                # action leaves position unchanged or reverses direction
                # so penalize distance by setting it to max possible dist
                d = self._dists[goal_coord][coord]
                d += 2
            else:
                d = self._dists[goal_coord][new_coord]
            dists.append(d)
        return dists



"""Policies for the PredatorPrey-v0 environment."""
import os.path as osp

from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

ENV_ID = "PredatorPrey-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "predator_prey")
policy_specs = {}

# PredatorPrey-10x10-P2-p3-s2-coop-v0
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "grid": "10x10",
            "num_predators": 2,
            "num_prey": 3,
            "cooperative": True,
            "prey_strength": 2,
            "obs_dim": 2,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=10x10-num_predators=2-num_prey=3-cooperative=True-prey_strength=2-obs_dim=2",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec


# PredatorPrey-10x10-P3-p3-s2-coop-v0
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "grid": "10x10",
            "num_predators": 3,
            "num_prey": 3,
            "cooperative": True,
            "prey_strength": 2,
            "obs_dim": 2,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=10x10-num_predators=3-num_prey=3-cooperative=True-prey_strength=2-obs_dim=2",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec


# PredatorPrey-10x10-P4-p3-s3-coop-v0
for policy_file_name in [
    "sp_seed0.pkl",
    "sp_seed1.pkl",
    "sp_seed2.pkl",
    "sp_seed3.pkl",
    "sp_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "grid": "10x10",
            "num_predators": 4,
            "num_prey": 3,
            "cooperative": True,
            "prey_strength": 3,
            "obs_dim": 2,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=10x10-num_predators=4-num_prey=3-cooperative=True-prey_strength=3-obs_dim=2",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec



"""Wrapper to flatten observations into 1D numpy arrays."""
from gymnasium import spaces

from posggym import Env, ObservationWrapper


class FlattenObservations(ObservationWrapper):
    """Observation wrapper that flattens the observation.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper

    Note
    ----
    This implementation is based on the similar Gymnasium wrapper:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/flatten_observation.py

    """

    def __init__(self, env: Env):
        super().__init__(env)
        self._observation_spaces = {
            i: spaces.flatten_space(obs_space)
            for i, obs_space in env.observation_spaces.items()
        }

    def observations(self, obs):
        return {
            i: spaces.flatten(self.env.observation_spaces[i], obs_i)
            for i, obs_i in obs.items()
        }



"""Policies for the Driving-v0 environment."""
import os.path as osp

from posggym.agents.torch_policy import PPOPolicy
from posggym.agents.utils import processors
from posggym.config import AGENT_MODEL_DIR

ENV_ID = "Driving-v0"
agent_model_dir = osp.join(AGENT_MODEL_DIR, "grid_world", "driving")
policy_specs = {}

# 7x7RoundAbout-n2-v0
for policy_file_name in [
    "klr_k0_seed0.pkl",
    "klr_k0_seed1.pkl",
    "klr_k0_seed2.pkl",
    "klr_k0_seed3.pkl",
    "klr_k0_seed4.pkl",
    "klr_k1_seed0.pkl",
    "klr_k1_seed1.pkl",
    "klr_k1_seed2.pkl",
    "klr_k1_seed3.pkl",
    "klr_k1_seed4.pkl",
    "klr_k2_seed0.pkl",
    "klr_k2_seed1.pkl",
    "klr_k2_seed2.pkl",
    "klr_k2_seed3.pkl",
    "klr_k2_seed4.pkl",
    "klr_k3_seed0.pkl",
    "klr_k3_seed1.pkl",
    "klr_k3_seed2.pkl",
    "klr_k3_seed3.pkl",
    "klr_k3_seed4.pkl",
    "klr_k4_seed0.pkl",
    "klr_k4_seed1.pkl",
    "klr_k4_seed2.pkl",
    "klr_k4_seed3.pkl",
    "klr_k4_seed4.pkl",
    "klrbr_k4_seed0.pkl",
    "klrbr_k4_seed1.pkl",
    "klrbr_k4_seed2.pkl",
    "klrbr_k4_seed3.pkl",
    "klrbr_k4_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "grid": "7x7RoundAbout",
            "num_agents": 2,
            "obs_dim": (3, 1, 1),
            "obstacle_collisions": False,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=7x7RoundAbout-num_agents=2-obs_dim=(3,1,1)-obstacle_collisions=False",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
    )
    policy_specs[spec.id] = spec


# 14x14RoundAbout-n2-v0
for policy_file_name in [
    "klr_k0_seed0.pkl",
    "klr_k0_seed1.pkl",
    "klr_k0_seed2.pkl",
    "klr_k0_seed3.pkl",
    "klr_k0_seed4.pkl",
    "klr_k1_seed0.pkl",
    "klr_k1_seed1.pkl",
    "klr_k1_seed2.pkl",
    "klr_k1_seed3.pkl",
    "klr_k1_seed4.pkl",
    "klr_k2_seed0.pkl",
    "klr_k2_seed1.pkl",
    "klr_k2_seed2.pkl",
    "klr_k2_seed3.pkl",
    "klr_k2_seed4.pkl",
    "klr_k3_seed0.pkl",
    "klr_k3_seed1.pkl",
    "klr_k3_seed2.pkl",
    "klr_k3_seed3.pkl",
    "klr_k3_seed4.pkl",
    "klr_k4_seed0.pkl",
    "klr_k4_seed1.pkl",
    "klr_k4_seed2.pkl",
    "klr_k4_seed3.pkl",
    "klr_k4_seed4.pkl",
    "klrbr_k4_seed0.pkl",
    "klrbr_k4_seed1.pkl",
    "klrbr_k4_seed2.pkl",
    "klrbr_k4_seed3.pkl",
    "klrbr_k4_seed4.pkl",
]:
    spec = PPOPolicy.get_spec_from_path(
        env_id=ENV_ID,
        env_args={
            "grid": "14x14RoundAbout",
            "num_agents": 2,
            "obs_dim": (3, 1, 1),
            "obstacle_collisions": False,
        },
        policy_file_path=osp.join(
            agent_model_dir,
            "grid=14x14RoundAbout-num_agents=2-obs_dim=(3,1,1)-obstacle_collisions=False",
            policy_file_name,
        ),
        version=0,
        valid_agent_ids=None,
        # policy is deterministic given random seed
        nondeterministic=False,
        # actions sampled, rather always taking most probable action
        deterministic=False,
        obs_processor_cls=processors.FlattenProcessor,
        obs_processor_config=None,
        action_processor_cls=None,
        action_processor_config=None,
    )
    policy_specs[spec.id] = spec



"""A passive environment checker wrapper for an environment."""
from typing import Dict

import posggym
import posggym.model as M
from posggym.utils.passive_env_checker import (
    check_agent_action_spaces,
    check_agent_observation_spaces,
    env_render_passive_checker,
    env_reset_passive_checker,
    env_step_passive_checker,
)


class PassiveEnvChecker(posggym.Wrapper):
    """A passive environment checker wrapper.

    Surrounds the step, reset and render functions to check that they follow the
    posggym environment and model APIs.

    Arguments
    ---------
    env : posggym.Env
        The environment to apply the wrapper

    Note
    ----
    This implementation is based on the similar Gymnasium wrapper:
    https://github.com/Farama-Foundation/Gymnasium/blob/v0.27.0/gymnasium/wrappers/env_checker.py

    """

    def __init__(self, env: posggym.Env):
        super().__init__(env)

        assert hasattr(env, "model"), "The environment must specify a model."

        assert hasattr(
            env, "action_spaces"
        ), "The environment must specify agent action spaces."
        check_agent_action_spaces(env.action_spaces)
        assert hasattr(
            env, "observation_spaces"
        ), "The environment must specify agent observation spaces."
        check_agent_observation_spaces(env.observation_spaces)

        self.checked_reset = False
        self.checked_step = False
        self.checked_render = False

    def step(self, actions: Dict[str, M.ActType]):
        """Steps through the environment.

        On the first call will run the `passive_env_step_check`.
        """
        if self.checked_step is False:
            self.checked_step = True
            return env_step_passive_checker(self.env, actions)
        else:
            return self.env.step(actions)  # type: ignore

    def reset(self, **kwargs):
        """Resets the environment.

        On the first call will run the `passive_env_reset_check`.

        """
        if self.checked_reset is False:
            self.checked_reset = True
            return env_reset_passive_checker(self.env, **kwargs)
        else:
            return self.env.reset(**kwargs)

    def render(self, *args, **kwargs):
        """Renders the environment.

        On the first call will run the `passive_env_render_check`.
        """
        if self.checked_render is False:
            self.checked_render = True
            return env_render_passive_checker(self.env, *args, **kwargs)
        else:
            return self.env.render(*args, **kwargs)



"""Functions and data structures for running experiments."""
import argparse
import json
import logging
import multiprocessing as mp
import os
import pathlib
import tempfile
import time
from datetime import datetime
from pprint import pformat
from typing import Any, Callable, Dict, List, NamedTuple, Optional, Tuple

import posggym
import posggym.agents.evaluation.render as render_lib
import posggym.agents.evaluation.stats as stats_lib
import posggym.agents.evaluation.writer as writer_lib
from posggym import wrappers
from posggym.agents import make
from posggym.agents.evaluation import runner
from posggym.config import BASE_RESULTS_DIR


LINE_BREAK = "-" * 60
EXP_ARG_FILE_NAME = "exp_args.json"


# A global lock used for controlling when processes print to stdout
# This helps keep top level stdout readable
LOCK = mp.Lock()


def _init_lock(lck):
    # pylint: disable=[global-statement]
    global LOCK
    LOCK = lck


class ExpParams(NamedTuple):
    """Params for a single experiment run."""

    exp_id: int
    env_id: str
    env_args: Optional[Dict[str, Any]]
    env_args_id: Optional[str]
    policy_ids: Dict[str, str]
    seed: int
    num_episodes: int
    time_limit: Optional[int] = None
    tracker_fn: Optional[Callable[[], List[stats_lib.Tracker]]] = None
    renderer_fn: Optional[Callable[[], List[render_lib.Renderer]]] = None
    render_mode: Optional[str] = None
    stream_log_level: int = logging.INFO
    file_log_level: int = logging.DEBUG
    record_env: bool = False
    # If None then uses the default cubic frequency
    record_env_freq: Optional[int] = None


def get_exp_parser() -> argparse.ArgumentParser:
    """Get command line argument parser with default experiment args."""
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )
    parser.add_argument(
        "--num_episodes",
        type=int,
        default=1000,
        help="Number of episodes per experiment.",
    )
    parser.add_argument(
        "--time_limit",
        type=int,
        default=None,
        help="Experiment time limit, in seconds.",
    )
    parser.add_argument(
        "--n_procs",
        type=int,
        default=1,
        help="Number of processors/experiments to run in parallel.",
    )
    parser.add_argument(
        "--log_level", type=int, default=21, help="Experiment log level."
    )
    parser.add_argument(
        "--root_save_dir",
        type=str,
        default=None,
        help=(
            "Optional directory to save results in. If supplied then it must "
            "be an existing directory. If None the default "
            "~/posggym_agents_results/<env_id>/results/ dir is used root "
            "results dir."
        ),
    )
    return parser


def make_exp_result_dir(
    exp_name: str,
    env_id: str,
    env_args_id: Optional[str],
    root_save_dir: Optional[str] = None,
) -> str:
    """Make a directory for experiment results."""
    time_str = datetime.today().strftime("%Y-%m-%d_%H-%M-%S")
    if root_save_dir is None and env_args_id is not None:
        root_save_dir = os.path.join(BASE_RESULTS_DIR, env_id, env_args_id)
    elif root_save_dir is None:
        root_save_dir = os.path.join(BASE_RESULTS_DIR, env_id)
    pathlib.Path(root_save_dir).mkdir(parents=True, exist_ok=True)
    result_dir = tempfile.mkdtemp(prefix=f"{exp_name}_{time_str}", dir=root_save_dir)
    return result_dir


def _log_exp_start(params: ExpParams, result_dir: str, logger: logging.Logger):
    LOCK.acquire()
    try:
        logger.info(LINE_BREAK)
        logger.info("Running with:")
        logger.info(pformat(params))
        logger.info(f"Result dir = {result_dir}")
        logger.info(LINE_BREAK)
    finally:
        LOCK.release()


def _log_exp_end(
    params: ExpParams, result_dir: str, logger: logging.Logger, exp_time: float
):
    LOCK.acquire()
    try:
        logger.info(LINE_BREAK)
        logger.info(f"Finished exp num {params.exp_id}")
        logger.info(f"Result dir = {result_dir}")
        logger.info(f"Experiment Run time {exp_time:.2f} seconds")
        logger.info(LINE_BREAK)
    finally:
        LOCK.release()


def get_exp_run_logger(
    exp_id: int,
    result_dir: str,
    stream_log_level: int = logging.INFO,
    file_log_level: int = logging.DEBUG,
) -> logging.Logger:
    """Get the logger for a single experiment run."""
    logger_name = f"exp_{exp_id}"
    logger = logging.getLogger(logger_name)
    logger.setLevel(min(stream_log_level, file_log_level))

    fname = f"exp_{exp_id}.log"
    log_file = os.path.join(result_dir, fname)
    file_formatter = logging.Formatter(
        # [Day-Month Hour-Minute-Second] exp_x Message
        "[%(asctime)s] %(levelname)s %(message)s",
        "%d-%m %H:%M:%S",
    )

    filehandler = logging.FileHandler(log_file)
    filehandler.setFormatter(file_formatter)
    filehandler.setLevel(file_log_level)

    stream_formatter = logging.Formatter(
        # [Day-Month Hour-Minute-Second] exp_x Message
        "[%(asctime)s] %(name)s %(message)s",
        "%d-%m %H:%M:%S",
    )
    streamhandler = logging.StreamHandler()
    streamhandler.setFormatter(stream_formatter)
    streamhandler.setLevel(stream_log_level)

    logger.addHandler(filehandler)
    logger.addHandler(streamhandler)
    logger.propagate = False

    return logger


def _get_exp_statistics(params: ExpParams) -> stats_lib.AgentStatisticsMap:
    stats = {}
    for i in params.policy_ids:
        stats[i] = {
            "exp_id": params.exp_id,
            "agent_id": i,
            "env_id": params.env_id,
            "env_args_id": params.env_args_id,
            "policy_id": params.policy_ids[i],
            "exp_seed": params.seed,
            "num_episodes": params.num_episodes,
            "time_limit": params.time_limit if params.time_limit else "None",
        }
    return stats


def _get_linear_episode_trigger(freq: int) -> Callable[[int], bool]:
    return lambda t: t % freq == 0


def run_single_experiment(args: Tuple[ExpParams, str]):
    """Run a single experiment and write results to a file."""
    params, result_dir = args
    exp_start_time = time.time()

    exp_logger = get_exp_run_logger(
        params.exp_id, result_dir, params.stream_log_level, params.file_log_level
    )
    _log_exp_start(params, result_dir, exp_logger)

    if params.env_args is not None:
        env = posggym.make(
            params.env_id, render_mode=params.render_mode, **params.env_args
        )
    else:
        env = posggym.make(params.env_id, render_mode=params.render_mode)
    assert len(params.policy_ids) == len(env.possible_agents), (
        f"Experiment env '{env}' has {len(env.possible_agents)} possible agents, but "
        f"only {len(params.policy_ids)} policies supplied."
    )

    if params.record_env:
        video_folder = os.path.join(result_dir, f"exp_{params.exp_id}_video")
        episode_trigger = None
        if params.record_env_freq:
            episode_trigger = _get_linear_episode_trigger(params.record_env_freq)
        env = wrappers.RecordVideo(env, video_folder, episode_trigger=episode_trigger)

    policies = {}
    for i, policy_id in params.policy_ids.items():
        policies[i] = make(policy_id, env.model, i)

    if params.tracker_fn:
        trackers = params.tracker_fn()
    else:
        trackers = stats_lib.get_default_trackers()

    renderers = params.renderer_fn() if params.renderer_fn else []
    writer = writer_lib.ExperimentWriter(
        params.exp_id, result_dir, _get_exp_statistics(params)
    )

    if params.seed is not None:
        env.reset(seed=params.seed)
        for idx, policy in enumerate(policies.values()):
            policy.reset(seed=params.seed + idx)

    try:
        statistics = runner.run_episode(
            env,
            policies,
            params.num_episodes,
            trackers,
            renderers,
            time_limit=params.time_limit,
            logger=exp_logger,
            writer=writer,
        )
        writer.write(statistics)

    except Exception as ex:
        exp_logger.exception("Exception occurred: %s", str(ex))
        exp_logger.error(pformat(locals()))
        raise ex
    finally:
        _log_exp_end(params, result_dir, exp_logger, time.time() - exp_start_time)
        env.close()
        for policy in policies.values():
            policy.close()
        for h in exp_logger.handlers:
            h.close()


def run_experiments(
    exp_name: str,
    exp_params_list: List[ExpParams],
    exp_log_level: int = logging.INFO + 1,
    n_procs: Optional[int] = None,
    exp_args: Optional[Dict] = None,
    root_save_dir: Optional[str] = None,
) -> str:
    """Run series of experiments.

    If exp_args is not None then will write to file in the result dir.
    """
    exp_start_time = time.time()
    logging.basicConfig(
        level=exp_log_level,
        # [Day-Month Hour-Minute-Second] Message
        format="[%(asctime)s] %(message)s",
        datefmt="%d-%m %H:%M:%S",
    )

    num_exps = len(exp_params_list)
    logging.log(exp_log_level, "Running %d experiments", num_exps)

    result_dir = make_exp_result_dir(
        exp_name,
        exp_params_list[0].env_id,
        exp_params_list[0].env_args_id,
        root_save_dir,
    )
    logging.log(exp_log_level, "Saving results to dir=%s", result_dir)

    if exp_args:
        write_experiment_arguments(exp_args, result_dir)

    if n_procs is None:
        n_procs = os.cpu_count()
    logging.log(exp_log_level, "Running %d processes", n_procs)

    mp_lock = mp.Lock()

    def _initializer(init_args):
        proc_lock = init_args
        _init_lock(proc_lock)

    if n_procs == 1:
        _initializer(mp_lock)
        for params in exp_params_list:
            run_single_experiment((params, result_dir))
    else:
        args_list = [(params, result_dir) for params in exp_params_list]
        with mp.Pool(n_procs, initializer=_initializer, initargs=(mp_lock,)) as p:
            p.map(run_single_experiment, args_list, 1)

    logging.log(exp_log_level, "Compiling results")
    writer_lib.compile_results(result_dir)

    logging.log(
        exp_log_level, "Experiment Run time %.2f seconds", time.time() - exp_start_time
    )

    return result_dir


def write_experiment_arguments(args: Dict[str, Any], result_dir: str) -> str:
    """Write experiment arguments to file."""
    arg_file = os.path.join(result_dir, EXP_ARG_FILE_NAME)
    with open(arg_file, "w", encoding="utf-8") as fout:
        json.dump(args, fout)
    return arg_file



"""Shortest path policies for pursuit evasion continuous environment."""
from __future__ import annotations

import math
from itertools import product
from typing import TYPE_CHECKING, List, Tuple, cast

import numpy as np

from posggym.agents.policy import Policy, PolicyID, PolicyState
from posggym.agents.utils import action_distributions
from posggym.envs.continuous.core import PMBodyState
from posggym.envs.continuous.pursuit_evasion_continuous import (
    PEAction,
    PEObs,
    PursuitEvasionContinuousModel,
)
from posggym.utils import seeding


if TYPE_CHECKING:
    from posggym.posggym.model import POSGModel
    from posggym.utils.history import AgentHistory


class PECShortestPathPolicy(Policy[PEAction, PEObs]):
    """Shortest path policy for pursuit evasion continuous environment."""

    def __init__(self, model: POSGModel, agent_id: str, policy_id: PolicyID):
        super().__init__(model, agent_id, policy_id)
        self.model = cast(PursuitEvasionContinuousModel, model)
        self._rng, _ = seeding.np_random()

        self.n_sensors = self.model.n_sensors
        self.sensor_obs_dim = self.model.sensor_obs_dim
        self.is_evader = self.agent_id == str(self.model.EVADER_IDX)
        self.action_space = self.model.action_spaces[self.agent_id]
        self.action_dtype = self.action_space.dtype

        # world without other agent for simulating next agent position
        self._world = self.model.world.copy()
        if self.is_evader:
            self._world.remove_entity("pursuer")
        else:
            self._world.remove_entity("evader")

        # all evader start and goal coords
        evader_coords = list(
            set(self._world.evader_start_coords + self._world.all_goal_coords)
        )
        self._dists = self._world.get_all_shortest_paths(evader_coords)

    def reset(self, *, seed: int | None = None):
        super().reset(seed=seed)
        if seed is not None:
            self._rng, _ = seeding.np_random(seed=seed)

    def get_initial_state(self) -> PolicyState:
        state = super().get_initial_state()
        state.update(
            {
                "action": None,
                "pi": None,
                "last_obs": None,
                "update_num": 0,
                # x, y, angle, vel_x, vel_y, vel_angle
                "body_state": PMBodyState(0, 0, 0, 0, 0, 0),
                "prev_body_state": PMBodyState(0, 0, 0, 0, 0, 0),
                "target_coord": np.zeros(2),
            }
        )
        return state

    def get_next_state(
        self,
        obs: PEObs,
        state: PolicyState,
    ) -> PolicyState:
        if state["update_num"] == 0:
            next_update_num = 1
            if self.is_evader:
                next_coord = obs[self.sensor_obs_dim + 1 : self.sensor_obs_dim + 3]
                target_coord = obs[self.sensor_obs_dim + 5 : self.sensor_obs_dim + 7]
            else:
                next_coord = obs[self.sensor_obs_dim + 3 : self.sensor_obs_dim + 5]
                target_coord = obs[self.sensor_obs_dim + 1 : self.sensor_obs_dim + 3]
            next_body_state = PMBodyState(next_coord[0], next_coord[1], 0, 0, 0, 0)
        else:
            next_update_num = state["update_num"] + 1
            next_body_state = self._get_next_body_state(
                state["body_state"], state["action"]
            )
            target_coord = state["target_coord"]

        _, pi = self._get_shortest_path_action(
            state["body_state"], next_body_state, target_coord
        )

        next_state = {
            "action": pi.sample(),
            "pi": pi,
            "last_obs": obs,
            "update_num": next_update_num,
            "body_state": next_body_state,
            "prev_body_state": state["body_state"],
            "target_coord": target_coord,
        }
        return next_state

    def get_state_from_history(self, history: AgentHistory) -> PolicyState:
        _, last_obs = history.get_last_step()
        if last_obs is None:
            return self.get_initial_state()

        state = self.get_initial_state()
        obs = history[0][1]
        if self.is_evader:
            next_coord = obs[self.sensor_obs_dim + 1 : self.sensor_obs_dim + 3]
            target_coord = obs[self.sensor_obs_dim + 5 : self.sensor_obs_dim + 7]
        else:
            next_coord = obs[self.sensor_obs_dim + 3 : self.sensor_obs_dim + 5]
            target_coord = obs[self.sensor_obs_dim + 1 : self.sensor_obs_dim + 3]
        state["last_obs"] = obs
        state["update_num"] = 1
        state["body_state"] = PMBodyState(next_coord[0], next_coord[1], 0, 0, 0, 0)
        state["target_coord"] = target_coord

        for a, o in history[1:]:
            state["last_obs"] = o
            state["update_num"] += 1
            state["body_state"] = self._get_next_body_state(state["body_state"], a)

        # need to get next action given final observation
        _, pi = self._get_shortest_path_action(state["body_state"], target_coord)

        state["action"] = np.array(pi.sample(), dtype=self.action_dtype)
        state["pi"] = pi
        return state

    def sample_action(self, state: PolicyState) -> PEAction:
        if state["pi"] is None:
            raise ValueError(
                "Policy state does not contain a valid action distribution. Make sure"
                "to call `step` or `get_next_state` before calling `sample_action`"
            )
        return np.array(state["pi"].sample(), dtype=self.action_dtype)

    def get_pi(self, state: PolicyState) -> action_distributions.ActionDistribution:
        return state["pi"]

    def get_value(self, state: PolicyState) -> float:
        raise NotImplementedError(
            f"`get_value()` no implemented by {self.__class__.__name__} policy"
        )

    def _get_next_body_state(
        self, body_state: PMBodyState, action: PEAction
    ) -> PMBodyState:
        agent_entity_id = "evader" if self.is_evader else "pursuer"

        self._world.set_entity_state(agent_entity_id, body_state)
        angle = body_state.angle + action[0]
        vel = self._world.linear_to_xy_velocity(action[1], angle)
        self._world.update_entity_state(agent_entity_id, angle=angle, vel=vel)

        self._world.simulate(1.0 / 10, 10)

        return self._world.get_entity_state(agent_entity_id)

    def _get_shortest_path_action(
        self,
        prev_body_state: PMBodyState,
        body_state: PMBodyState,
        target_coord: np.ndarray,
    ) -> Tuple[List[PEAction], action_distributions.ActionDistribution]:
        angle_vels = [
            -self.model.dyaw_limit,
            -self.model.dyaw_limit / 2.0,
            0.0,
            self.model.dyaw_limit / 2.0,
            self.model.dyaw_limit,
        ]
        linear_vels = [0.0, 0.5, 1.0]

        target_int_coord = self._world.convert_to_coord(target_coord)

        # find closest unblocked int coord to agent's float coord
        agent_coord = (body_state.x, body_state.y)
        agent_int_coord = self._world.convert_to_coord(agent_coord)
        dist = self._world.euclidean_dist(agent_coord, agent_int_coord)
        for x, y in product(
            [math.floor(body_state.x), math.ceil(body_state.x)],
            [math.floor(body_state.y), math.ceil(body_state.y)],
        ):
            if (x, y) in self._dists[target_int_coord]:
                d = self._world.euclidean_dist(agent_coord, (x, y))
                if d < dist:
                    agent_int_coord = (x, y)
                    dist = d

        # find adjacent int coords that are on shortest path to target
        target_next_agent_int_coords = [agent_int_coord]
        sp_dist = self._dists[target_int_coord][agent_int_coord]
        for dxy in [(-1, 0), (1, 0), (0, -1), (0, 1)]:
            next_agent_int_coord = (
                agent_int_coord[0] + dxy[0],
                agent_int_coord[1] + dxy[1],
            )
            if next_agent_int_coord not in self._dists[target_int_coord]:
                continue

            dist = self._dists[target_int_coord][next_agent_int_coord]
            if dist == sp_dist:
                target_next_agent_int_coords.append(next_agent_int_coord)
            elif dist < sp_dist:
                target_next_agent_int_coords = [next_agent_int_coord]
                sp_dist = dist
        target_next_agent_int_coord = self._rng.choice(target_next_agent_int_coords)

        # find actions that move agent towards target adjacent coord
        sp_actions = [(-self.model.dyaw_limit, 0.0)]
        sp_dist = float("inf")
        for a in product(angle_vels, linear_vels):
            next_body_state = self._get_next_body_state(body_state, a)
            dist = self._world.euclidean_dist(
                (next_body_state.x, next_body_state.y), target_next_agent_int_coord
            )
            if np.isclose(dist, sp_dist):
                sp_actions.append(a)
            elif dist < sp_dist:
                sp_dist = dist
                sp_actions = [a]

        # Need to handle case where agent gets stuck
        if np.isclose(
            (prev_body_state.x, prev_body_state.y),
            (body_state.x, body_state.y),
            rtol=0.0,
            atol=1e-1,
        ).all():
            if all(a[1] == 0.0 for a in sp_actions):
                # try to move forward
                sp_actions = [(a[0], a[1] + 0.5) for a in sp_actions]
            elif all(a[0] == 0.0 for a in sp_actions):
                # try to turn
                old_sp_actions = sp_actions
                sp_actions = []
                for a in old_sp_actions:
                    sp_actions += [
                        (a[0] + self.model.dyaw_limit / 2.0, a[1]),
                        (a[0] - self.model.dyaw_limit / 2.0, a[1]),
                    ]

        pi = action_distributions.DiscreteActionDistribution(
            {a: 1.0 / len(sp_actions) for a in sp_actions},
            self._rng,
        )
        return sp_actions, pi



"""Classes and functions for recording agent statistics during episodes."""
import abc
import time
from collections import ChainMap
from typing import Any, Dict, Iterable, List, Sequence

import numpy as np

import posggym
import posggym.model as M
from posggym.agents.policy import Policy

AgentStatisticsMap = Dict[str, Dict[str, Any]]


def generate_episode_statistics(trackers: Iterable["Tracker"]) -> AgentStatisticsMap:
    """Generate episode statistics from set of trackers."""
    statistics = combine_statistics([t.get_episode() for t in trackers])
    return statistics


def generate_statistics(trackers: Iterable["Tracker"]) -> AgentStatisticsMap:
    """Generate summary statistics from set of trackers."""
    statistics = combine_statistics([t.get() for t in trackers])
    return statistics


def combine_statistics(
    statistic_maps: Sequence[AgentStatisticsMap],
) -> AgentStatisticsMap:
    """Combine multiple Agent statistic maps into a single one."""
    agent_ids = list(statistic_maps[0].keys())
    return {
        i: dict(ChainMap(*(stat_maps[i] for stat_maps in statistic_maps)))
        for i in agent_ids
    }


def get_default_trackers() -> List["Tracker"]:
    """Get the default set of Trackers."""
    return [EpisodeTracker()]


class Tracker(abc.ABC):
    """Generic Tracker Base class."""

    @abc.abstractmethod
    def step(
        self,
        episode_t: int,
        env: posggym.Env,
        timestep: M.JointTimestep,
        action: Dict[str, M.ActType],
        policies: Dict[str, Policy],
        episode_end: bool,
    ):
        """Accumulates statistics for a single step."""

    @abc.abstractmethod
    def reset(self):
        """Reset all gathered statistics."""

    @abc.abstractmethod
    def reset_episode(self):
        """Reset all statistics prior to each episode."""

    @abc.abstractmethod
    def get_episode(self) -> AgentStatisticsMap:
        """Aggregate episode statistics for each agent."""

    @abc.abstractmethod
    def get(self) -> AgentStatisticsMap:
        """Aggregate all episode statistics for each agent."""


class EpisodeTracker(Tracker):
    """Tracks episode return and other statistics."""

    def __init__(self):
        # initialized when step first called
        self._agents = None

        self._num_episodes = 0
        self._current_start_time = time.time()
        # is initialized when step is first called
        self._current_returns = None
        self._current_dones = None
        self._current_steps = None
        self._current_outcomes = None

        self._dones = []
        self._times = []
        self._returns = []
        self._steps = []
        self._outcomes = []

    def step(
        self,
        episode_t: int,
        env: posggym.Env,
        timestep: M.JointTimestep,
        action: Dict[str, M.ActType],
        policies: Dict[str, Policy],
        episode_end: bool,
    ):
        if self._agents is None:
            self._agents = env.possible_agents

        if self._current_returns is None:
            self._current_returns = {i: 0.0 for i in env.possible_agents}

        if self._current_dones is None:
            self._current_dones = {i: False for i in env.possible_agents}

        if self._current_steps is None:
            self._current_steps = {i: 0 for i in env.possible_agents}

        if self._current_outcomes is None:
            self._current_outcomes = {i: M.Outcome.NA for i in env.possible_agents}

        if episode_t == 0:
            return

        for i in env.possible_agents:
            self._current_returns[i] += timestep.rewards.get(i, 0.0)

            if i in timestep.observations and not self._current_dones[i]:
                self._current_steps[i] += 1

            if not self._current_dones[i] and (
                timestep.terminations.get(i, False)
                or timestep.truncations.get(i, False)
            ):
                self._current_dones[i] = True
                self._current_outcomes[i] = timestep.infos.get("outcome", M.Outcome.NA)

        if episode_end:
            self._num_episodes += 1
            self._times.append(time.time() - self._current_start_time)
            self._returns.append(self._current_returns)
            self._dones.append(self._current_dones)
            self._steps.append(self._current_steps)

            outcome = {}
            for i in env.possible_agents:
                if i in timestep.infos:
                    outcome[i] = timestep.infos.get("outcome", M.Outcome.NA)
                else:
                    outcome[i] = M.Outcome.NA

                if outcome[i] == M.Outcome.NA:
                    # in case outcome was received earlier in episode (i.e. agent
                    # finished early)
                    outcome[i] = self._current_outcomes[i]
            self._outcomes.append(outcome)

    def reset(self):
        self.reset_episode()
        self._num_episodes = 0
        self._dones = []
        self._times = []
        self._returns = []
        self._steps = []
        self._outcomes = []

    def reset_episode(self):
        self._current_done = False
        self._current_start_time = time.time()

        if self._current_returns is not None:
            for i in self._current_returns:
                self._current_returns[i] = 0.0

        if self._current_dones is not None:
            for i in self._current_dones:
                self._current_dones[i] = False

        if self._current_steps is not None:
            for i in self._current_steps:
                self._current_steps[i] = 0

        if self._current_outcomes is not None:
            for i in self._current_outcomes:
                self._current_outcomes[i] = M.Outcome.NA

    def get_episode(self) -> AgentStatisticsMap:
        assert self._agents is not None

        stats = {}
        for i in self._agents:
            stats[i] = {
                "episode_number": self._num_episodes,
                "episode_return": self._returns[-1][i],
                "episode_steps": self._steps[-1][i],
                "episode_outcome": self._outcomes[-1][i],
                "episode_done": self._dones[-1][i],
                "episode_time": self._times[-1],
            }

        return stats

    def get(self) -> AgentStatisticsMap:
        assert self._agents is not None

        outcome_counts = {k: {i: 0 for i in self._agents} for k in M.Outcome}
        for outcome in self._outcomes:
            for i in self._agents:
                outcome_counts[outcome[i]][i] += 1

        stats = {}
        for i in self._agents:
            returns_i = [ep_return[i] for ep_return in self._returns]
            steps_i = [ep_steps[i] for ep_steps in self._steps]
            dones_i = [ep_done[i] for ep_done in self._dones]
            stats[i] = {
                "num_episodes": self._num_episodes,
                "episode_return_mean": np.mean(returns_i, axis=0),
                "episode_return_std": np.std(returns_i, axis=0),
                "episode_return_max": np.max(returns_i, axis=0),
                "episode_return_min": np.min(returns_i, axis=0),
                "episode_steps_mean": np.mean(steps_i),
                "episode_steps_std": np.std(steps_i),
                "episode_time_mean": np.mean(self._times),
                "episode_time_std": np.std(self._times),
                "num_episode_done": np.sum(dones_i),
            }

            for outcome, counts in outcome_counts.items():
                stats[i][f"num_{outcome}"] = counts[i]

        return stats
